'''Implement persistent storage of any objects.'''

import os
import pathlib
import pickle

from typing import Optional

SUFFIX = '.pickle'

SlotId = int


def _slot_id_to_prefix(idx: SlotId) -> str:
    return f'{idx:03}'


class Store:
    '''Implement storage of objects into "slots" that are just numbered
    files on the disk. This class keeps track of a "current" slot,
    which is updated when slots are loaded, saved and deleted.

    - Objects are always saved into a new slot, and this new slot
      becomes the current one.
    - Only current slot can be deleted, and there is no current slot
      anymore after the delete operation.
    - Load operations makes the loaded slot the current one.
    '''
    def _scan(self):
        self._files = sorted(self._base_dir.rglob(f'*{SUFFIX}'))
        self._slot_ids = [int(f.name[:-len(SUFFIX)]) for f in self._files]
        assert len(self._files) == len(self._slot_ids)
        # We need to compute the next available slot id that we
        # can use to save an object. Since slots can be deleted,
        # we find the max SlotId currently used and just add 1, or
        # start at 0 if no slot is in used.
        if len(self._slot_ids) <= 0:
            self._next_slot_id = 0
        else:
            self._next_slot_id = max(self._slot_ids) + 1

    def __init__(self, base_dir: str):
        self._base_dir = pathlib.Path(base_dir)
        self._files = []
        self._slot_ids = []
        self._cur_idx = -1
        self._scan()

    def destroy(self) -> None:
        '''Destroy all slots and data in storage'''
        for path in self._files:
            path.unlink()
        if self._base_dir.exists():
            self._base_dir.rmdir()
        self._scan()

    def save(self, obj) -> SlotId:
        '''Saves the given object in a new slot'''
        self._base_dir.mkdir(parents=True, exist_ok=True)
        saved = self._next_slot_id
        prefix = _slot_id_to_prefix(saved)
        path = self._base_dir.joinpath(f'{prefix}{SUFFIX}')
        with path.open("wb") as fout:
            pickle.dump(obj, fout)
        self._scan()
        count = len(self._files)
        assert count > 0
        # Saving an object makes the slot it was saved into the
        # current one. Since we always save in the max slot id, which
        # should always be associated with the last index
        self._cur_idx = count - 1
        return saved

    def delete(self) -> Optional[SlotId]:
        '''Deletes the current slot'''
        index_to_delete = self._cur_idx
        if index_to_delete < 0:
            return None
        slot_id = self._slot_ids[index_to_delete]
        os.remove(self._files[index_to_delete])
        self._scan()
        self._cur_idx = -1
        return slot_id

    def can_load(self) -> bool:
        '''Returns whether we have anything to load'''
        return len(self._files) > 0

    def load(self, goto_next: bool):
        '''Loads the next or previous object'''
        count = len(self._files)
        assert count > 0
        inc = 1 if goto_next else -1
        if self._cur_idx < 0:
            self._cur_idx = count - 1 if goto_next else 0
        self._cur_idx = (self._cur_idx + inc) % count
        path = self._files[self._cur_idx]
        slot_id = self._slot_ids[self._cur_idx]
        with path.open("rb") as fin:
            return pickle.load(fin), slot_id



'''Module that wraps curses color support into higher level palettes.'''

import curses

from . import types

Pair = int
PairId = int
ColorId = int

# The first 64 color pair ids are reserved to render SGR
# https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
SGR_COLOR_PAIR_ID_COUNT = 64

# Status bar color pair id and background color
BAR_COLOR_PAIR_ID = SGR_COLOR_PAIR_ID_COUNT
BAR_COLOR_BG = 39  # 249

# First color pair id reserved for filters. The actual number of color reserved
# depends on the number of colors in the palette.
FIRST_FILTER_COLOR_PAIR_ID = BAR_COLOR_PAIR_ID + 1

# https://stackoverflow.com/questions/18551558/how-to-use-terminal-color-palette-with-curses
PALETTES = [
    # Dark theme, "error" in red first
    [
        196,  # Red
        208,  # Orange
        190,  # Yellow
        46,   # Green
        33,   # Blue
        201,  # Purple
        219,  # Pink
    ],
    # Light theme, "error" in red first
    [
        1,    # Red
        208,  # Orange
        3,    # Yellow
        2,    # Green
        12,   # Blue
        129,  # Purple
        201,  # Pink
    ],
    # Dark theme, "ok" in green first
    [
        46,   # Green
        190,  # Yellow
        208,  # Orange
        196,  # Red
        33,   # Blue
        201,  # Purple
        219,  # Pink
    ],
    # Light theme, "ok" in green first
    [
        2,    # Green
        3,    # Yellow
        208,  # Orange
        1,    # Red
        12,   # Blue
        129,  # Purple
        201,  # Pink
    ],
    # Dark theme, "neutral"
    [
        33,   # Blue
        201,  # Purple
        219,  # Pink
        190,  # Yellow
        46,   # Green
        208,  # Orange
        196,  # Red
    ],
    # Light theme, "neutral"
    [
        12,   # Blue
        129,  # Purple
        201,  # Pink
        3,    # Yellow
        2,    # Green
        208,  # Orange
        1,    # Red
    ],
]


def cycle_palette(
        palette_id: types.PaletteId,
        forward: bool,
) -> types.PaletteId:
    '''Returns the palette index after the given one in the given direction.'''
    assert palette_id >= 0
    assert palette_id < len(PALETTES)
    incr = 1 if forward else -1
    return (palette_id + incr) % len(PALETTES)


def apply_palette(
        palette_id: types.PaletteId,
        reverse: bool,
) -> None:
    '''Applies given palette to curses.'''
    # Note: python raises IndexError for us if palette_index is out of range
    pal = PALETTES[palette_id]
    for i, color in enumerate(pal):
        pair_id = FIRST_FILTER_COLOR_PAIR_ID + i
        if reverse:
            curses.init_pair(pair_id, 0, color)
        else:
            curses.init_pair(pair_id, color, -1)


def get_color_pair(
        palette_id: types.PaletteId,
        filter_index: int,
) -> Pair:
    '''Gets the curses.color_pair associated with given palette and filter.'''
    if filter_index < 0:
        return curses.color_pair(0)
    colors_count = len(PALETTES[palette_id])
    pair_id = FIRST_FILTER_COLOR_PAIR_ID + (filter_index % colors_count)
    return curses.color_pair(pair_id)


def get_color_pair_id(
        fg: ColorId,
        bg: ColorId,
) -> PairId:
    '''Gets color pair associated with given foreground and background.'''
    assert -1 <= fg <= 15, fg
    assert -1 <= bg <= 15, bg
    assert bg != -1 or fg != -1
    if bg == -1:
        return fg
    if fg == -1:
        return 16 + bg
    if fg == bg:
        return 32 + fg
    return 48 + bg


def get_fg_bg_color_pair(
        fg: ColorId,
        bg: ColorId
) -> Pair:
    '''Returns the color pair use to display the given fg and bg colors.'''
    return curses.color_pair(get_color_pair_id(fg, bg))


def init_color_pairs():
    '''Initiliazes the color pairs we depend on.'''
    curses.init_pair(BAR_COLOR_PAIR_ID, 0, BAR_COLOR_BG)
    for i in range(16):
        curses.init_pair(i, i, -1)
        curses.init_pair(i + 16, -1, i)
        curses.init_pair(i + 32, i, i)
        curses.init_pair(i + 48, 15, i)


def init():
    '''Initializes color support.'''
    assert curses.has_colors()
    assert curses.COLORS >= 256, 'Not enough colors (try TERM=screen-256color)'
    curses.start_color()
    curses.use_default_colors()
    init_color_pairs()



'''Unit tests for colors'''

# To properly test colors and palette support, or visualize existing
# palette definition, run python3 -m searchf.test.color

import os
import curses

from .. import colors

TERM = os.environ.get('TERM', None)


def _setup():
    print(f'TERM={TERM}')
    if TERM is None:
        return False
    curses.initscr()
    # curses will not have COLORS attribute when running within pytest,
    # so we artificially provide one
    curses.COLORS = 256
    return True


def test_init():
    '''Test colors.init()'''
    if not _setup():
        return
    # WARNING: this is not a real test since function has side-effects
    # we are not validating
    colors.init()


def test_cycle_palette():
    '''Test colors.cycle_palette()'''
    max_index = len(colors.PALETTES) - 1
    assert colors.cycle_palette(0, True) == 1
    assert colors.cycle_palette(0, False) == max_index
    assert colors.cycle_palette(max_index, True) == 0
    assert colors.cycle_palette(max_index, False) == max_index - 1


def test_apply_palette():
    '''Tests colors.apply_palette()'''
    if not _setup():
        return
    # WARNING: this is not a real test since function has side-effects
    # we are not validating
    for pid in range(len(colors.PALETTES)):
        colors.apply_palette(pid, False)
        colors.apply_palette(pid, True)


def test_get_color_pair():
    '''Tests colors.get_color_pair()'''
    if not _setup():
        return
    # WARNING: this is not a real test since function has side-effects
    # we are not validating
    colors.get_color_pair(0, -1)
    colors.get_color_pair(0, 0)


def test_get_fg_bg_color_pair():
    '''Tests colors.get_fg_bg_color_pair()'''
    if not _setup():
        return
    colors.get_fg_bg_color_pair(-1, 0)



'''Application module'''

import copy
import curses
import curses.ascii
import os
import sys

from curses.textpad import Textbox
from typing import Any
from typing import List
from typing import Tuple

from . import debug
from . import enums
from . import keys
from . import types
from . import views

# Changes layout to show a debug window in which debug() function will output
USE_DEBUG = False

STATUS_UNCHANGED = 'unchanged'


def load_lines(path: str) -> List[str]:
    '''Load all the lines of the given file.'''
    with open(path, encoding='utf-8') as f:
        # splitlines() ensures that any newline char is removed
        return f.read().splitlines()


def get_max_yx(scr) -> types.Size:
    '''This function is a test artifact that wraps getmaxyx() from curses so
    that we can overwrite it and test specific dimensions.
    '''
    return scr.getmaxyx()


def getmtime(path):
    '''Wraps os.getmtime() for testing.'''
    return os.path.getmtime(path)


class EscapeException(Exception):
    '''Signals that Escape key has been pressed.'''


def validate(k: int) -> int:
    '''Validates key.'''
    if k == curses.ascii.DEL:
        k = curses.KEY_BACKSPACE
    elif k == curses.ascii.ESC:
        raise EscapeException()
    return k


def get_text(scr, y: int, x: int, text_prompt: str, handler, text: str) -> str:
    '''Gets text interactively from end user.'''
    _, maxw = get_max_yx(scr)
    scr.addstr(y, x, text_prompt)
    x += len(text_prompt)
    width = max(0, maxw - x)
    editwin = curses.newwin(1, width, y, x)
    scr.refresh()
    box = Textbox(editwin)
    for c in text:
        box.do_command(ord(c))
    text = ''
    try:
        handler(box)
        text = box.gather().strip()
    except EscapeException:
        pass

    editwin.clear()
    editwin.refresh()
    clear(scr, y, 0, len(text_prompt))
    return text if text else ''


def prompt(scr, y: int, x: int, text_prompt: str, text: str) -> str:
    '''Prompts user to enter some text.'''
    def handle(box):
        box.edit(validate=validate)
    return get_text(scr, y, x, text_prompt, handle, text)


def clear(scr, y: int, x: int, length: int):
    '''Prints "length" spaces at the given position.'''
    _, maxw = get_max_yx(scr)
    blank = f'{" ":>{length}}'
    scr.addstr(y, x, blank[:maxw-(x+1)])
    scr.move(y, x)


class App:
    '''Aggregates all views, controlling which ones are visible, handling
    key presses and command dispatching.'''

    # pylint: disable=too-many-instance-attributes

    def __init__(self, help_lines: List[str]) -> None:
        self._help_lines: List[str] = help_lines
        self._debug_view: views.DebugView
        self._event_view: views.KeyEventView
        self._scr: Any = None
        self._margins: types.Margins = types.Margins()
        self._show_events: bool = False
        self._path: str = ''
        self._views: List[views.TextView] = []
        self._help_view_index: int = -1
        self._current: int = -1
        self._hidden_view: int = -1
        self._y_get_text: int = 0
        self._mtime: float = 0.0
        self._auto_reload: bool = False
        self._auto_reload_anchor: int = 0

    def layout(self) -> types.Size:
        '''Recompute layout'''
        assert self._scr
        scr = self._scr
        maxh, maxw = get_max_yx(scr)
        maxh -= (self._margins.top + self._margins.bottom)
        maxw -= (self._margins.left + self._margins.right)
        x = self._margins.left
        y = self._margins.top

        if self._show_events:
            self._event_view.layout(y, x, 2, maxw)
            y += 2
            maxh -= 2

        view_lines_count = maxh

        if USE_DEBUG:
            dbg_size = (10, maxw)
            self._debug_view.layout(dbg_size, (y, x))
            # Just add padding to expose layout issues in the app
            padding = 3
            maxh = max(0, maxh - (2 * padding + dbg_size[0]))
            maxw = max(0, maxw - 2 * padding)
            view_lines_count = maxh
            x = padding
            y = dbg_size[0] + padding

        for v in self._views:
            v.layout(view_lines_count, maxw, y, x)

        y += view_lines_count
        self._y_get_text = y
        return maxh, maxw

    def _set_view(self, idx: int, propagate_config: bool) -> types.Status:
        assert 0 <= idx < len(self._views)
        if self._current == idx:
            return f'Current view is already {self._views[idx].name()}'
        self._hidden_view = \
            self._current if idx == self._help_view_index else -1
        config = self._views[self._current].get_config()
        self._current = idx
        if propagate_config:
            self._views[idx].set_config(copy.deepcopy(config))
        self._views[idx].show()
        return f'Switched to {self._views[idx].name()}'

    def _help_view_push(self) -> types.Status:
        return self._set_view(self._help_view_index, False)

    def _help_view_pop(self) -> types.Status:
        assert self._hidden_view >= 0
        return self._set_view(self._hidden_view, False)

    def _reload(self, scroll_to: int) -> types.Status:
        lines = load_lines(self._path)
        for i, v in enumerate(self._views):
            if i == self._help_view_index:
                v.set_lines(self._help_lines)
            else:
                v.set_lines(lines)
                v.set_v_offset(scroll_to, False)
        self._views[self._current].draw()
        self._mtime = getmtime(self._path)
        return 'File reloaded'

    def create(
            self,
            store,
            scr,
            margins: types.Margins,
            show_events: bool,
            path: str,
    ) -> None:
        '''Creates all views in the given screen, and loads the content from
        the given file.'''
        self._scr = scr
        self._margins = margins
        self._show_events = show_events
        self._path = path
        self._views.clear()
        self._views.append(views.TextView(store, scr, 'View 1', path))
        self._views.append(views.TextView(store, scr, 'View 2', path))
        self._views.append(views.TextView(store, scr, 'View 3', path))
        self._views.append(views.TextView(store, scr, 'Help', 'Help'))
        self._help_view_index = len(self._views) - 1
        self._views[self._help_view_index].get_config().line_numbers = False
        if USE_DEBUG:
            self._debug_view = views.DebugView(scr)
            debug.OUT_FUNC = self._debug_view.out
        self._event_view = views.KeyEventView(scr)
        self.layout()
        self._reload(0)
        self._set_view(0, False)

    def prompt(self, text_prompt: str, text: str) -> str:
        '''Prompts user to enter or edit some text.'''
        return prompt(self._scr,
                      self._y_get_text,
                      self._margins.left,
                      text_prompt,
                      text)

    def _get_keyword(self) -> str:
        return self.prompt('Keyword: ', '')

    def try_start_search(self) -> types.Status:
        '''Tries to initiate a less like search.'''
        v = self._views[self._current]
        if v.has_filters():
            return 'Cannot start search when filters already defined'
        t = self._get_keyword()
        return v.search(t)

    def _toggle_auto_reload(self, anchor: int) -> types.Status:
        self._auto_reload = not self._auto_reload
        self._auto_reload_anchor = anchor
        return f'Auto reload {self._auto_reload}'

    def _poll(self) -> Tuple[bool, types.Status]:
        if self._auto_reload:
            mtime = getmtime(self._path)
            if mtime != self._mtime:
                return True, self._reload(self._auto_reload_anchor)
        return False, STATUS_UNCHANGED

    def handle_event(
            self,
            ev: keys.KeyEvent,
    ) -> Tuple[bool, types.Status]:
        '''Handles the given key, propagating it to the proper view.'''

        if self._show_events:
            self._event_view.show(ev)

        if ev.is_poll():
            return self._poll()

        # Local functions redirecting to current view v
        v = self._views[self._current]

        def new_keyword(new_filter) -> types.Status:
            keyword = self._get_keyword()
            return v.push_keyword(keyword, new_filter)

        def goto_line() -> types.Status:
            line_as_text = self.prompt('Enter line: ', '')
            return v.goto_line(line_as_text)

        def edit_keyword() -> types.Status:
            count, keyword = v.get_last_keyword()
            if count <= 0:
                return 'No keyword to edit'
            keyword = self.prompt('Edit: ', keyword if keyword else '')
            if len(keyword) <= 0:
                return 'No change made'
            v.execute(enums.Command.POP_KEYWORD)
            v.push_keyword(keyword, count == 1)
            return 'Keyword updated'

        def resize() -> types.Status:
            self._scr.clear()
            self._scr.refresh()
            size = self.layout()
            v.draw()
            return f'Resized to {size[1]}x{size[0]}'

        # Map the commands requiring custom functions and that
        # cannot be directly sent over the current view.
        cmd_to_func = {
            enums.Command.SHOW_VIEW_1:
                lambda: self._set_view(0, False),
            enums.Command.SHOW_VIEW_2:
                lambda: self._set_view(1, False),
            enums.Command.SHOW_VIEW_3:
                lambda: self._set_view(2, False),
            enums.Command.SHOW_VIEW_1_WITH_FILTER:
                lambda: self._set_view(0, True),
            enums.Command.SHOW_VIEW_2_WITH_FILTER:
                lambda: self._set_view(1, True),
            enums.Command.SHOW_VIEW_3_WITH_FILTER:
                lambda: self._set_view(2, True),
            enums.Command.SHOW_HELP:
                self._help_view_push,
            enums.Command.EDIT_KEYWORD:
                edit_keyword,
            enums.Command.PUSH_KEYWORD:
                lambda: new_keyword(False),
            enums.Command.PUSH_FILTER_AND_KEYWORD:
                lambda: new_keyword(True),
            enums.Command.RELOAD_HEAD:
                lambda: self._reload(0),
            enums.Command.RELOAD_HEAD_AUTO:
                lambda: self._toggle_auto_reload(0),
            enums.Command.RELOAD_TAIL:
                lambda: self._reload(sys.maxsize),
            enums.Command.RELOAD_TAIL_AUTO:
                lambda: self._toggle_auto_reload(sys.maxsize),
            enums.Command.TRY_SEARCH:
                self.try_start_search,
            enums.Command.GOTO_LINE:
                goto_line,
            enums.Command.RESIZE:
                resize,
        }

        handled: bool = True  # Assumed until otherwise
        status: types.Status = ''
        if ev.cmd == enums.Command.QUIT or ev.key == curses.ascii.ESC:
            if self._hidden_view < 0:
                handled = False
            else:
                status = self._help_view_pop()
        elif ev.cmd in cmd_to_func:
            status = cmd_to_func[ev.cmd]()
        elif ev.cmd:
            status = v.execute(ev.cmd)
        else:
            handled = False
            status = f'Unknown key {ev.text} (? for help, q to quit)'

        return handled, status



'''Unit tests for models'''

from .. import enums
from .. import models


def test_filter():
    '''Test models.Filter'''
    f = models.Filter()
    f.add('Keyword')

    count, keyword = f.get_count_and_last_keyword()
    assert count == 1
    assert keyword == 'Keyword'

    f.pop()
    count, keyword = f.get_count_and_last_keyword()
    assert count == 0
    assert not keyword


def test_digit_count():
    '''Test models.digit_count()'''
    assert models.digits_count(0) == 1
    assert models.digits_count(9) == 1
    assert models.digits_count(10) == 2
    assert models.digits_count(99) == 2
    assert models.digits_count(123456) == 6


SGR_MODE = enums.SgrMode.REMOVE


def test_model():
    '''Test models.RawContent'''
    rc = models.RawContent()
    assert rc.line_count() == 0

    rc.set_lines([])
    assert rc.line_count() == 0
    assert rc.line_number_length() == 1

    all_lines = [
        'A very simple first line',
        'Another line',
        'And a third one',
        'A fourth',
        'And finally the very last one']
    rc.set_lines(all_lines)
    assert rc.line_count() == len(all_lines)
    assert rc.line_number_length() == 1

    sc = rc.filter([], enums.LineVisibility.ALL, SGR_MODE)
    assert sc.hits_count() == 0

    f = models.Filter()
    f.add('not match')
    sc = rc.filter([f], enums.LineVisibility.ALL, SGR_MODE)
    assert sc.hits_count() == 0

    f.pop()
    f.add('simple')
    sc = rc.filter([f], enums.LineVisibility.ALL, SGR_MODE)
    assert sc.hits_count() == 1
    assert sc.visible_line_count() == len(all_lines)
    sc = rc.filter([f], enums.LineVisibility.CONTEXT_1, SGR_MODE)
    assert sc.visible_line_count() == 2
    sc = rc.filter([f], enums.LineVisibility.ONLY_MATCHING, SGR_MODE)
    assert sc.visible_line_count() == 1

    f.pop()
    f.add('line')
    sc = rc.filter([f], enums.LineVisibility.ALL, SGR_MODE)
    assert sc.hits_count() == 2
    assert sc.visible_line_count() == len(all_lines)
    sc = rc.filter([f], enums.LineVisibility.ONLY_MATCHING, SGR_MODE)
    assert sc.visible_line_count() == 2

    f.pop()
    f.add('Another')
    sc = rc.filter([f], enums.LineVisibility.ALL, SGR_MODE)
    assert sc.hits_count() == 1
    assert sc.visible_line_count() == len(all_lines)
    sc = rc.filter([f], enums.LineVisibility.CONTEXT_1, SGR_MODE)
    assert sc.visible_line_count() == 3
    sc = rc.filter([f], enums.LineVisibility.CONTEXT_2, SGR_MODE)
    assert sc.visible_line_count() == 4
    sc = rc.filter([f], enums.LineVisibility.CONTEXT_5, SGR_MODE)
    assert sc.visible_line_count() == 5
    sc = rc.filter([f], enums.LineVisibility.ONLY_MATCHING, SGR_MODE)
    assert sc.visible_line_count() == 1

    # Add a ruler
    f.pop()
    f.add('very')
    sc = rc.filter([f], enums.LineVisibility.CONTEXT_1, SGR_MODE)
    assert sc.hits_count() == 2


def test_display_content():
    '''Test models.DisplayContent'''
    dc = models.DisplayContent()

    rc = models.RawContent()
    rc.set_lines([
        'A very simple first line',
        'Another line',
        'And a third one'])
    sc = rc.filter([], enums.LineVisibility.ALL, SGR_MODE)

    dc = sc.layout(1, 1, True)
    assert dc
    dc = sc.layout(1, 1, False)
    assert dc

    f = models.Filter()
    f.add('third')
    sc = rc.filter([f], enums.LineVisibility.CONTEXT_1, SGR_MODE)
    dc = sc.layout(1, 1, True)
    assert dc


def test_offsets():
    '''Test models.Offsets'''

    co = models.Offsets()
    co.layout(10, 100)

    co.set_h_offset(0)
    co.set_h_offset(100)

    co.set_v_offset(0)
    assert co.voffset == 0
    co.set_v_offset(1)
    assert co.voffset == 1
    co.set_v_offset(200)
    assert co.voffset == 90


def test_view_config():
    '''Test models.ViewConfig'''
    vc = models.ViewConfig()
    assert not vc.has_filters()

    last = models.Filter()
    vc.push_filter(last)
    assert vc.has_filters()
    assert vc.get_filters_count() == 1
    assert last == vc.top_filter()

    last = models.Filter()
    vc.push_filter(last)
    assert vc.get_filters_count() == 2
    assert last == vc.top_filter()
    vc.swap_filters(0, 1)
    assert last != vc.top_filter()

    vc.rotate_filters(False)
    assert last == vc.top_filter()
    vc.rotate_filters(True)
    assert last != vc.top_filter()

    vc.set_palette(1)



'''https://github.com/human3/searchf'''

import os
import argparse
import curses

from . import __version__
from . import __url__
from . import app
from . import colors
from . import enums
from . import keys
from . import storage
from . import types
from . import utils


def _load_help_lines():
    lines = [f'  ~ Searchf {__version__} Help ~', '']
    help_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'help.txt')
    lines += app.load_lines(help_file)
    lines += ['More info']
    lines += [f'  {__url__}']
    lines += [f'  {__url__}/releases/tag/{__version__}']
    return lines


HELP_LINES = _load_help_lines()

APP = app.App(HELP_LINES)


class StatusView:
    '''View class for a single status line at the bottom of the screen.'''
    def __init__(self, scr):
        self.pos = (0, 0)
        self._scr = scr
        self._width = 0

    def layout(self) -> None:
        '''Layout the view.'''
        max_y, max_x = app.get_max_yx(self._scr)
        # We use padding to workaround addstr() throwing exception
        # when printing outside of the screen boundaries with
        # some utf-8 encoded char
        padding = 5
        self._width = max(0, max_x - 1 - 2 * padding)
        y = max_y - 1
        x = min(padding, max_x - 1)
        self.pos = (y, x)

    def draw(self, status) -> None:
        '''Draw the status.'''
        pos = self.pos
        self._scr.addstr(pos[0], pos[1], f'{status:^{self._width}}')
        self._scr.refresh()


def main_loop(scr,
              path: str,
              use_debug: bool,
              show_events: bool,
              keys_processor: keys.Processor
              ) -> None:
    '''Main loop consuming keys and events.'''
    app.USE_DEBUG = use_debug
    colors.init()
    scr.refresh()  # Must be call once on empty screen?
    curses.mousemask(-1)
    store = storage.Store('.searchf')
    margins = types.Margins()
    margins.bottom += 1
    APP.create(store, scr, margins, show_events, path)
    v = StatusView(scr)
    v.layout()

    scr.timeout(200)
    status = ''
    while True:
        scr.move(v.pos[0], 0)
        try:
            ev = keys_processor.get()
        except KeyboardInterrupt:
            break
        handled, new_status = APP.handle_event(ev)
        if not handled and ev.cmd == enums.Command.QUIT:
            break
        if new_status == app.STATUS_UNCHANGED:
            continue
        if ev.cmd == enums.Command.RESIZE:
            v.layout()
        app.clear(scr, v.pos[0], v.pos[1], len(status))
        status = new_status
        v.draw(status)


def main_curses(scr, args) -> None:
    '''Main entry point requiring curse environment.'''
    main_loop(scr,
              args.file,
              args.debug,
              args.show_events,
              keys.Processor(scr, curses))


def init_env() -> argparse.ArgumentParser:
    '''Initialize environment and return argument parser.'''
    # https://stackoverflow.com/questions/27372068/why-does-the-escape-key-have-a-delay-in-python-curses
    os.environ.setdefault('ESCDELAY', '25')
    os.environ['TERM'] = 'screen-256color'
    parser = argparse.ArgumentParser(
        description='Console application to search into text files and \
highlight keywords.',
        epilog='Press ? in the application for more information, or go to \
https://github.com/human3/searchf')
    parser.add_argument('file')
    parser.add_argument('--debug',
                        help='Use debug layout',
                        action='store_true')
    parser.add_argument('--show-events',
                        help='Show events like key presses',
                        action='store_true')
    return parser


def main() -> None:
    '''Application entry point.'''
    parser = init_env()
    args = parser.parse_args()
    utils.wrapper(False, curses.wrapper, main_curses, args)


if __name__ == '__main__':  # pragma: no cover
    main()



'''Module abstracting key input.'''

from typing import Any
from typing import List
from typing import Optional

import datetime
import curses
import curses.ascii
import sys

from . import enums

# Special key that will result in app being polled
POLL = -1
UNMAP = -2

# Number of seconds to wait before canceling key escape sequence
ESCAPE_TIMEOUT = 0.1


# Map keys to simple text view commands
KEYS_TO_COMMAND = {
    ord('F'):              enums.Command.POP_FILTER,
    curses.KEY_BACKSPACE:  enums.Command.POP_FILTER,
    curses.ascii.DEL:      enums.Command.POP_FILTER,
    ord('-'):              enums.Command.POP_KEYWORD,
    ord('_'):              enums.Command.POP_KEYWORD,
    ord('n'):              enums.Command.VSCROLL_TO_NEXT_MATCH,
    ord('p'):              enums.Command.VSCROLL_TO_PREV_MATCH,
    ord('l'):              enums.Command.TOGGLE_LINE_NUMBERS,
    ord('k'):              enums.Command.TOGGLE_WRAP,
    ord('*'):              enums.Command.TOGGLE_BULLETS,
    ord('.'):              enums.Command.TOGGLE_SHOW_SPACES,
    ord('`'):              enums.Command.NEXT_SGR_MODE,
    ord('~'):              enums.Command.PREV_SGR_MODE,
    ord('c'):              enums.Command.NEXT_PALETTE,
    ord('C'):              enums.Command.PREV_PALETTE,
    ord('h'):              enums.Command.NEXT_COLORIZE_MODE,
    ord('H'):              enums.Command.PREV_COLORIZE_MODE,
    ord('m'):              enums.Command.NEXT_LINE_VISIBILITY,
    ord('M'):              enums.Command.PREV_LINE_VISIBILITY,
    ord('i'):              enums.Command.TOGGLE_IGNORE_CASE,
    ord('x'):              enums.Command.TOGGLE_HIDING,
    curses.KEY_UP:         enums.Command.GO_UP,
    curses.KEY_DOWN:       enums.Command.GO_DOWN,
    curses.KEY_LEFT:       enums.Command.GO_LEFT,
    curses.KEY_RIGHT:      enums.Command.GO_RIGHT,
    curses.KEY_HOME:       enums.Command.GO_HOME,
    ord('g'):              enums.Command.GO_HOME,
    ord('<'):              enums.Command.GO_HOME,
    curses.KEY_END:        enums.Command.GO_END,
    ord('G'):              enums.Command.GO_END,
    ord('>'):              enums.Command.GO_END,
    curses.KEY_NPAGE:      enums.Command.GO_NPAGE,
    ord(' '):              enums.Command.GO_NPAGE,
    curses.KEY_PPAGE:      enums.Command.GO_PPAGE,
    ord('b'):              enums.Command.GO_PPAGE,
    curses.KEY_SLEFT:      enums.Command.GO_SLEFT,
    curses.KEY_SRIGHT:     enums.Command.GO_SRIGHT,
    ord('d'):              enums.Command.SWAP_FILTERS,
    ord('w'):              enums.Command.ROTATE_FILTERS_UP,
    ord('s'):              enums.Command.ROTATE_FILTERS_DOWN,
    ord('\\'):             enums.Command.SLOT_SAVE,
    ord('|'):              enums.Command.SLOT_DELETE,
    ord(']'):              enums.Command.SLOT_LOAD_NEXT,
    ord('['):              enums.Command.SLOT_LOAD_PREV,
    ord('1'):              enums.Command.SHOW_VIEW_1,
    ord('2'):              enums.Command.SHOW_VIEW_2,
    ord('3'):              enums.Command.SHOW_VIEW_3,
    ord('!'):              enums.Command.SHOW_VIEW_1_WITH_FILTER,
    ord('@'):              enums.Command.SHOW_VIEW_2_WITH_FILTER,
    ord('#'):              enums.Command.SHOW_VIEW_3_WITH_FILTER,
    ord('?'):              enums.Command.SHOW_HELP,
    ord('e'):              enums.Command.EDIT_KEYWORD,
    ord('+'):              enums.Command.PUSH_KEYWORD,
    ord('='):              enums.Command.PUSH_KEYWORD,
    ord('f'):              enums.Command.PUSH_FILTER_AND_KEYWORD,
    ord('\n'):             enums.Command.PUSH_FILTER_AND_KEYWORD,
    ord('r'):              enums.Command.RELOAD_HEAD,
    ord('R'):              enums.Command.RELOAD_HEAD_AUTO,
    ord('t'):              enums.Command.RELOAD_TAIL,
    ord('T'):              enums.Command.RELOAD_TAIL_AUTO,
    ord('/'):              enums.Command.TRY_SEARCH,
    curses.ascii.TAB:      enums.Command.GOTO_LINE,
    curses.ascii.BEL:      enums.Command.GOTO_LINE,
    ord('q'):              enums.Command.QUIT,
    ord('Q'):              enums.Command.QUIT,
    curses.KEY_RESIZE:     enums.Command.RESIZE,
}

MOUSE_STATE_TO_CMD = {}

MOUSE_STATE_TO_CMD[curses.BUTTON4_PRESSED] = enums.Command.GO_UP
if getattr(curses, 'BUTTON5_PRESSED', None):
    MOUSE_STATE_TO_CMD[curses.BUTTON5_PRESSED] = enums.Command.GO_DOWN
    # If BUTTON5_PRESSED is not defined, we need to workaround it see:
    # https://github.com/peterbrittain/asciimatics/issues/345
    # https://github.com/python/cpython/issues/91132
elif sys.platform == 'darwin':
    # mac use old curses (see https://github.com/python/cpython/issues/91132)
    MOUSE_STATE_TO_CMD[0x8000000] = enums.Command.GO_DOWN
else:
    # we assume PDCurses
    # https://github.com/mattn/pdcurses/blob/master/curses.h#L194
    MOUSE_STATE_TO_CMD[0x200000] = enums.Command.GO_DOWN


KEYS_TO_TEXT = {
    POLL:                 'POLL',
    UNMAP:                'UNMAP',
    curses.ascii.ESC:     'ESC',
    curses.ascii.DEL:     'DELETE',
    ord(' '):             'SPACE',
    ord('\n'):            'ENTER',
    ord('\t'):            'TAB',
    curses.KEY_BACKSPACE: 'BACKSPACE',
    curses.KEY_UP:        'UP',
    curses.KEY_DOWN:      'DOWN',
    curses.KEY_LEFT:      'LEFT',
    curses.KEY_RIGHT:     'RIGHT',
    curses.KEY_HOME:      'HOME',
    curses.KEY_END:       'END',
    curses.KEY_PPAGE:     'PGUP',
    curses.KEY_NPAGE:     'PGDOWN',
    curses.KEY_MOUSE:     'MOUSE',
    curses.KEY_RESIZE:    'RESIZE',
}


def _key_to_text(key: int) -> str:
    if key in KEYS_TO_TEXT:
        return KEYS_TO_TEXT[key]
    return f'{chr(key)}'


class KeyEvent():
    '''Class representing KeyEvent '''
    key: int
    text: str
    cmd: Optional[enums.Command]

    def is_poll(self) -> bool:
        '''Returns whether or not this is the poll key/command'''
        return self.key == POLL

    def __init__(self,
                 key: int = POLL,
                 text: str = '',
                 cmd: Optional[enums.Command] = None):
        self.key = key
        self.text = text
        self.cmd = cmd
        if not cmd and key in KEYS_TO_COMMAND:
            self.cmd = KEYS_TO_COMMAND[key]
        if not text or len(text) <= 0:
            self.text = _key_to_text(key)


class Provider:
    '''Key press provider, that replaces curses getch()
    implementation when testing.'''
    def __init__(self, keys: List[Any]) -> None:
        self._keys: List[Any] = keys

    def getch(self) -> int:
        '''Gets the next key'''
        key = self._keys.pop(0)
        return key if isinstance(key, int) else ord(key)


# Define strings used for representing key combination
SHIFT = '[1;2'
ALT = '[1;3'
ALT_SHIFT = '[1;4'
CTRL = '[1;5'
CTRL_SHIFT = '[1;6'

UP = 'A'
DOWN = 'B'
RIGHT = 'C'
LEFT = 'D'


# Some OS are eating up lots of dead key + arrow combinations
ESCAPED_TO_COMMAND = {
    chr(curses.KEY_UP):    enums.Command.ROTATE_FILTERS_UP,
    CTRL + UP:             enums.Command.ROTATE_FILTERS_UP,
    CTRL_SHIFT + UP:       enums.Command.ROTATE_FILTERS_UP,
    ALT + UP:              enums.Command.ROTATE_FILTERS_UP,
    chr(curses.KEY_DOWN):  enums.Command.ROTATE_FILTERS_DOWN,
    CTRL + DOWN:           enums.Command.ROTATE_FILTERS_DOWN,
    ALT + DOWN:            enums.Command.ROTATE_FILTERS_DOWN,
    CTRL_SHIFT + DOWN:     enums.Command.ROTATE_FILTERS_DOWN,
    chr(curses.KEY_LEFT):  enums.Command.SWAP_FILTERS,
    CTRL + LEFT:           enums.Command.SWAP_FILTERS,
    CTRL_SHIFT + LEFT:     enums.Command.SWAP_FILTERS,
    ALT + LEFT:            enums.Command.SWAP_FILTERS,
    chr(curses.KEY_RIGHT): enums.Command.SWAP_FILTERS,
    CTRL + RIGHT:          enums.Command.SWAP_FILTERS,
    CTRL_SHIFT + RIGHT:    enums.Command.SWAP_FILTERS,
    ALT + RIGHT:           enums.Command.SWAP_FILTERS,
    SHIFT + UP:            enums.Command.GO_PPAGE,
    SHIFT + DOWN:          enums.Command.GO_NPAGE,
    SHIFT + LEFT:          enums.Command.GO_SLEFT,
    SHIFT + RIGHT:         enums.Command.GO_SRIGHT,
    'OH':                  enums.Command.GO_HOME,
    'OF':                  enums.Command.GO_END,
}


class Processor:
    '''Class that returns key pressed by end-user, handling any required
    decoding of escape sequences.
    '''
    def __init__(self, getch_provider, getmouse_provider) -> None:
        getch_attr = getattr(getch_provider, 'getch', None)
        assert getch_attr
        assert callable(getch_attr)
        self._getch_provider = getch_provider
        getmouse_attr = getattr(getmouse_provider, 'getmouse', None)
        assert getmouse_attr
        assert callable(getmouse_attr)
        self._getmouse_provider = getmouse_provider
        self._escaping = False
        self._seq = ''
        self.start_: datetime.datetime = datetime.datetime.min

    def _start_esc(self) -> None:
        self.start_ = datetime.datetime.now()
        self._escaping = True
        self._seq = ''

    def _stop_esc(self) -> str:
        seq = self._seq
        self._escaping = False
        self._seq = ''
        return seq

    def process(self, key: int) -> KeyEvent:
        '''Process given key'''
        # pylint: disable=too-many-return-statements

        if not self._escaping:
            if key == curses.KEY_MOUSE:
                state = self._getmouse_provider.getmouse()[4]
                if state in MOUSE_STATE_TO_CMD:
                    return KeyEvent(key, '', MOUSE_STATE_TO_CMD[state])
                key = POLL
            elif key == curses.ascii.ESC:
                self._start_esc()
                key = POLL
            return KeyEvent(key)

        assert self._escaping
        if key == curses.ascii.ESC:
            # Trash unrecognized sequence, and start again
            self._start_esc()
            return KeyEvent(POLL)
        if key < 0:
            # Assume sequence timed out
            seq = self._stop_esc()
            if len(seq) <= 0:
                # Emtpy escaped sequence, we remit the ESC key we swallowed
                key = curses.ascii.ESC
            else:
                # The sequence we captured is unrecognised
                key = UNMAP
            return KeyEvent(key, seq, None)
        delta = datetime.datetime.now() - self.start_
        if delta.total_seconds() > ESCAPE_TIMEOUT:
            # Stop escaping regardless of sequence after 200ms
            seq = self._stop_esc()
            return KeyEvent(key)

        # Ok to store key in sequence
        self._seq += chr(key)
        # debug.out(f'{key} {chr(key)} {str(key)} {delta.total_seconds()}')
        if self._seq not in ESCAPED_TO_COMMAND:
            return KeyEvent(POLL)
        seq = self._stop_esc()
        cmd = ESCAPED_TO_COMMAND[seq]
        return KeyEvent(key, seq, cmd)

    def get(self) -> KeyEvent:
        '''Wait on next key event typed by user or -1 if none'''
        return self.process(self._getch_provider.getch())



'''Module exporting basic types'''

from dataclasses import dataclass
from typing import Tuple
from typing import NamedTuple


@dataclass
class Margins():
    '''Defining margins for window.'''
    top: int
    bottom: int
    left: int
    right: int

    def __init__(self):
        self.top = 0
        self.bottom = 0
        self.left = 0
        self.right = 0


class Position(NamedTuple):
    '''Defining text cursor position'''
    x: int
    y: int


Size = Tuple[int, int]
Status = str
PaletteId = int



'''Interactive test of colors used by searchf'''

import curses
import sys
from .. import colors

# https://stackoverflow.com/questions/18551558/how-to-use-terminal-color-palette-with-curses

DEFAULT_PALETTE = range(16)
DEFAULT_PALETTE_EXT = range(16, 32)


def _reset():
    for i in range(curses.COLORS-1):
        curses.init_pair(i, i, -1)
    for i in range(16):
        curses.init_pair(i + 16, -1, i)
        curses.init_pair(i + 32, i, i)
        curses.init_pair(i + 48, 15, i)


def _show_all(scr):
    scr.move(0, 0)
    scr.addstr('Flat list of color:\n')
    for i in range(8):
        fgc, bgc = curses.pair_content(i)
        scr.addstr(f'{i} {fgc} {bgc} ', curses.color_pair(i))
    scr.addstr('\n')
    for i in range(curses.COLORS-1):
        scr.addstr(f'{i}', curses.color_pair(i))
    scr.addstr('\n')


def _show_palette(scr, name, pal):
    '''Draw sample text on screen using the given palette'''
    scr.addstr(f'{name:12}')
    for _, color in enumerate(pal):
        scr.addstr(f' {color:<3}', curses.color_pair(color))
    scr.addstr('\n')


def _wait(scr):
    scr.addstr('\n<Press q to quit, any other key to continue>\n')
    scr.refresh()
    key = scr.getch()
    if key == ord('q'):
        sys.exit(0)


def _show_all_and_wait(scr):
    _show_all(scr)
    _wait(scr)


def _show_all_palettes(scr):
    _reset()
    _show_all_and_wait(scr)
    for i, _ in enumerate(colors.PALETTES):
        colors.apply_palette(i, False)
        _show_all_and_wait(scr)

        colors.apply_palette(i, True)
        _show_all_and_wait(scr)


def _one_screen(scr):
    _reset()
    scr.move(0, 0)
    name = 'Palette/idx'
    scr.addstr(f'{name:12}')
    for i in range(16):
        scr.addstr(f' {i:<3}')
    scr.addstr('\n')

    for i, pal in enumerate(colors.PALETTES):
        _show_palette(scr, f'{i}', pal)

    _show_palette(scr, 'default', DEFAULT_PALETTE)
    _show_palette(scr, '', DEFAULT_PALETTE_EXT)
    _wait(scr)


def main(scr):
    '''Module entry point'''
    assert curses.has_colors()
    colors.init()
    _show_all_palettes(scr)
    _one_screen(scr)


curses.wrapper(main)



'''This module contains helper functions to define and manipulate segments.
'''

import re

from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Iterable
from typing import Set
from typing import Tuple


class Segment(NamedTuple):
    '''A segment is a pair of indices (start, end) that identifies a
    portion of text displayed with a given style. Please note that
    start is inclusive, while end is exclusive.
    '''
    start: int
    end: int
    attr: int  # Filter index if < 256, or raw curses attributes


def merge(bottom: List[Segment], top: List[Segment]) -> List[Segment]:
    '''Merge 2 layers of segment into one single layer. All segments of the
    top layer remain on top and visible, while segments of the bottom layer
    can be partially or completely obstructed by overlapping segments from
    the top layer.

    '''
    merged: List[Segment] = []

    # The logic here is to primarily iterate over each bottom segment and then
    # consume all the top segments (if any) that are overlapping the current
    # bottom segment.
    i_b = 0
    i_t = 0
    while i_b < len(bottom):
        assert bottom[i_b].start < bottom[i_b].end
        if i_t >= len(top):
            # we consumed all tops, so just add all the bottoms
            merged.append(bottom[i_b])
            i_b += 1
            continue
        assert top[i_t].start < top[i_t].end
        if top[i_t].end <= bottom[i_b].start:
            # top is clear ahead
            merged.append(top[i_t])
            i_t += 1
            continue
        if top[i_t].start >= bottom[i_b].end:
            # bottom is clear ahead
            merged.append(bottom[i_b])
            i_b += 1
            continue
        # we have some overlap
        bot = bottom[i_b]
        i_bnext = bot.start
        while i_bnext < bot.end:
            if i_t >= len(top):
                # we consumed all tops, so just add last segment of bottom
                if i_bnext < bot.end:
                    merged.append(Segment(i_bnext, bot.end, bot.attr))
                break
            assert top[i_t].start < top[i_t].end
            if top[i_t].start <= i_bnext:
                merged.append(top[i_t])
                i_bnext = top[i_t].end
                i_t += 1
                continue
            end = min(top[i_t].start, bot.end)
            merged.append(Segment(i_bnext, end, bot.attr))
            i_bnext = end
        i_b += 1

    assert i_b >= len(bottom)

    while i_t < len(top):
        merged.append(top[i_t])
        i_t += 1

    return merged


def flatten(layers: List[List[Segment]]) -> List[Segment]:
    '''Flatten the array of layers into one single layers.'''
    assert len(layers) > 0
    merged = layers[0]
    for i in range(1, len(layers)):
        merged = merge(merged, layers[i])
    return merged


def sort_and_merge(segments: Set[Segment]) -> List[Segment]:
    '''Merges overlapping segments belonging to the same filter, sort them, and
    returns a clean sorted list of non overlapping segments. For instance, in
    "abcde" the 2 keywords "abc" and "cde" are repectively matching segments
    (0,3) and (2,5) which are overlapping. This function would merge them as
    one segment (0,5).
    '''
    merged: Set[Segment] = set()
    pending: Optional[Segment] = None
    for cur in sorted(segments):
        assert cur.start < cur.end
        if not pending:
            pending = cur
        elif pending.end >= cur.start:
            assert pending.attr == cur.attr
            pending = Segment(
                pending.start, max(pending.end, cur.end), cur.attr)
        else:
            merged.add(pending)
            pending = cur
    if pending:
        merged.add(pending)
    return sorted(merged)


def find_matching(
        text: str,
        keywords,
        ignore_case: bool,
        attr: int,
        ) -> Tuple[bool, List[Segment]]:
    '''Returns True and the segments matching any of the given keywords if
    each keyword is found at least once in the given text, False and
    an empty list otherwise.
    '''
    assert len(keywords) > 0
    flags = re.IGNORECASE if ignore_case else 0
    matching: Set[Segment] = set()
    # matching is a set() as multiple matches are possible
    for k in keywords:
        is_matching = False
        for match in re.finditer(k, text, flags):
            if match.start() >= match.end():
                continue
            is_matching = True
            matching.add(Segment(match.start(), match.end(), attr))
        if not is_matching:
            # Bail out early as soon as one keyword has no match
            return False, []

    # Sort all segments and then merge them as overlap can happen
    return True, sort_and_merge(matching)


def iterate(
        start: int,
        end: int,
        matching_segments: List[Segment]
        ) -> Iterable[Tuple[bool, int, int, int]]:
    '''This function is used to build the list of text draw commands to
    display a line containing highlighted keywords. Text draws
    commands are yielded for left to right drawing, alternating
    highlighted segments (prefixed by True) with normal ones (prefixed
    by False).

    The arguments start and end define the boundaries of the text we
    care about, and constrain all the returned draw commands. The
    given 'matching_segments' define the segments that must be
    highlighted.
    '''
    assert start < end
    for cur in matching_segments:
        assert cur.start < cur.end  # Paranoid overkill assert

        if start >= end:
            # start has moved past the area of interest: could
            # break out but we are done for good so return!
            return

        # Consume "s", the current matching segment, by inspecting
        # its relative position with (start, end) segment.

        if start >= cur.end:
            # The current matching segment is of no interest, but we
            # need to keep iterating to find next matching segment of
            # interest to the right (if any)
            continue
        if end < cur.start:
            # Current matching segment and all remaining ones (on the
            # right) are of no interest. (start, end) defines a non
            # matching segment that we yield after breaking out of the
            # for loop
            break
        if start < cur.start:
            # There is a non-matching segment before s that we must
            # yield now
            yield (False, start, cur.start, -1)
            start = cur.start
        # Take care of matching segment within s (if any)
        matching_end = min(cur.end, end)
        if start < matching_end:
            yield (True, start, matching_end, cur.attr)
        # s has been entirely consumed
        start = cur.end

    # We walked through all matching_segments (if any), we now need to
    # address any non-matching segment that has been left out
    if start < end:
        yield (False, start, end, -1)



'''Provides enum helper and enums for application.'''

from enum import auto
from enum import Enum


class AutoEnum(Enum):
    '''Base class for auto enum that can get iterated over with
    wrapping.'''
    def __new__(cls, description):
        value = len(cls.__members__)
        obj = object.__new__(cls)
        obj._value_ = value
        obj._description = description
        return obj

    def __str__(self):
        return f'{self._description}'  # f'{self._value_} {self._description}'

    @classmethod
    def from_int(cls, i):
        '''Returns the enum member associated with the given integer.'''
        for member in cls:
            if member.value == i:
                return member
        raise ValueError('Unsupported enum value')

    @classmethod
    def get(cls, i):
        '''Returns enumeration matching given int value'''
        i = i % len(cls.__members__)
        return cls.from_int(i)

    def get_next(self):
        '''Returns the next enumeration.'''
        return self.__class__.get(self.value + 1)

    def get_prev(self):
        '''Returns the previous enumeraion.'''
        return self.__class__.get(self.value - 1)


class LineVisibility(AutoEnum):
    '''Line visibility modes.'''
    ONLY_MATCHING = 'Showing only matching lines'
    CONTEXT_1 = 'Reveal 1 line above/below matching'
    CONTEXT_2 = 'Reveal 2 lines above/below matching'
    CONTEXT_5 = 'Reveal 5 lines above/below matching'
    ALL = 'Showing all lines'


class ColorizeMode(AutoEnum):
    '''Keyword colorize modes.'''
    KEYWORD_HIGHLIGHT = 'Keyword highlight'
    KEYWORD = 'Keyword'
    LINE = 'Line'


class SgrMode(AutoEnum):
    '''Select Graphic Rendition processing modes.'''
    PROCESS = 'Process SGR (ie colorize)'
    REMOVE = 'Remove SGR (ie do not colorize)'
    NONE = 'Do not process SGR (ie passthrough)'


class Command(Enum):
    '''Commands supported by application, and mostly triggered by
    end-user's key presses.'''
    QUIT = auto()
    GO_UP = auto()
    GO_DOWN = auto()
    GO_LEFT = auto()
    GO_RIGHT = auto()
    GO_HOME = auto()
    GO_END = auto()
    GO_NPAGE = auto()
    GO_PPAGE = auto()
    GO_SLEFT = auto()
    GO_SRIGHT = auto()
    POP_FILTER = auto()
    POP_KEYWORD = auto()
    VSCROLL_TO_NEXT_MATCH = auto()
    VSCROLL_TO_PREV_MATCH = auto()
    TOGGLE_LINE_NUMBERS = auto()
    TOGGLE_WRAP = auto()
    TOGGLE_BULLETS = auto()
    TOGGLE_SHOW_SPACES = auto()
    NEXT_SGR_MODE = auto()
    PREV_SGR_MODE = auto()
    NEXT_COLORIZE_MODE = auto()
    PREV_COLORIZE_MODE = auto()
    NEXT_LINE_VISIBILITY = auto()
    PREV_LINE_VISIBILITY = auto()
    TOGGLE_IGNORE_CASE = auto()
    TOGGLE_HIDING = auto()
    NEXT_PALETTE = auto()
    PREV_PALETTE = auto()
    SWAP_FILTERS = auto()
    ROTATE_FILTERS_UP = auto()
    ROTATE_FILTERS_DOWN = auto()
    SLOT_SAVE = auto()
    SLOT_DELETE = auto()
    SLOT_LOAD_NEXT = auto()
    SLOT_LOAD_PREV = auto()
    SHOW_VIEW_1 = auto()
    SHOW_VIEW_2 = auto()
    SHOW_VIEW_3 = auto()
    SHOW_VIEW_1_WITH_FILTER = auto()
    SHOW_VIEW_2_WITH_FILTER = auto()
    SHOW_VIEW_3_WITH_FILTER = auto()
    SHOW_HELP = auto()
    PUSH_KEYWORD = auto()
    EDIT_KEYWORD = auto()
    PUSH_FILTER_AND_KEYWORD = auto()
    RELOAD_HEAD = auto()
    RELOAD_HEAD_AUTO = auto()
    RELOAD_TAIL = auto()
    RELOAD_TAIL_AUTO = auto()
    TRY_SEARCH = auto()
    GOTO_LINE = auto()
    RESIZE = auto()



'''Unit tests for segments'''

from .. import enums
from .. import colors
from .. import sgr

# When running test with pytest, curses is not initialized and actual
# color pairs are not available. We just pretend that color pair id
# are actual valid color pair (both types resolving to int)
sgr.FG_BG_TO_COLOR_PAIR_FUNC = colors.get_color_pair_id


def test_processor():
    '''Test sgr.Processor()'''
    p = sgr.Processor()

    line = 'Some dummy line'
    a, new_line = p.filter(line, enums.SgrMode.NONE)
    assert len(a) == 0
    assert new_line == line
    a, new_line = p.filter(line, enums.SgrMode.REMOVE)
    assert len(a) == 0
    assert new_line == line
    a, new_line = p.filter(line, enums.SgrMode.PROCESS)
    assert len(a) == 0
    assert new_line == line

    line = '\x1B[30m30 \
Black\x1B[0m \
\x1B[30;1mbold\x1B[0m \
\x1B[30;2mdim\x1B[0m \
\x1B[30;4munderline\x1B[0m \
\x1B[40mreverse\x1B[0m \
\x1B[40;97mreverse\x1B[0m \
\x1B[40;30mreverse\x1B[0m'
    line_bare = '30 Black bold dim underline reverse reverse reverse'

    a, new_line = p.filter(line, enums.SgrMode.NONE)
    assert len(a) == 0
    assert new_line == line
    a, new_line = p.filter(line, enums.SgrMode.REMOVE)
    assert len(a) == 0
    assert new_line == line_bare
    a, new_line = p.filter(line, enums.SgrMode.PROCESS)
    assert len(a) == 6, a
    assert new_line == line_bare

    # Attributes carry over properly accross multiple lines
    a, new_line = p.filter('\x1B[31mSomething red', enums.SgrMode.PROCESS)
    assert len(a) == 1
    a, new_line = p.filter('still red\x1B[0m', enums.SgrMode.PROCESS)
    assert len(a) == 1



'''Unit tests for segments'''

from .. import segments


def _iterate(start, end, test_segments, expected_segments):
    actual_segments = list(segments.iterate(start, end, test_segments))
    assert len(actual_segments) == len(expected_segments)
    for actual, expected in zip(actual_segments, expected_segments):
        _, start, end, _ = actual
        assert start < end, f'{start} {end} not a segment!'
        assert actual == expected, f'{actual} != {expected}'


def test_iterate():
    '''Test segments.iterate()'''
    _iterate(0, 20,
             {segments.Segment(11, 12, -1), segments.Segment(20, 21, -1)},
             [(False, 0, 11, -1), (True, 11, 12, -1), (False, 12, 20, -1)])
    seg = {segments.Segment(5, 15, -1), segments.Segment(25, 35, -1)}
    _iterate(0, 50, seg,
             [(False, 0, 5, -1), (True, 5, 15, -1), (False, 15, 25, -1),
              (True, 25, 35, -1), (False, 35, 50, -1)])
    _iterate(10, 20, seg,
             [(True, 10, 15, -1), (False, 15, 20, -1)])
    _iterate(0, 4, seg,
             [(False, 0, 4, -1)])
    _iterate(7, 12, seg,
             [(True, 7, 12, -1)])
    _iterate(17, 20, seg,
             [(False, 17, 20, -1)])
    _iterate(10, 50, seg,
             [(True, 10, 15, -1), (False, 15, 25, -1), (True, 25, 35, -1),
              (False, 35, 50, -1)])
    _iterate(0, 10, {segments.Segment(5, 15, -1)},
             [(False, 0, 5, -1), (True, 5, 10, -1)])


def _sort_and_merge(test_segments, expected_segments):
    count = 0
    for actual, expected in zip(
            segments.sort_and_merge(test_segments), expected_segments):
        assert actual[0] < actual[1]
        assert actual == expected, f'{actual} != {expected}'
        count += 1
    assert count == len(expected_segments)


def test_sort_and_merge():
    '''Test segments.sort_and_merge()'''
    seg = {segments.Segment(0, 1, -1), segments.Segment(2, 3, -1)}
    _sort_and_merge(seg, seg)
    _sort_and_merge({segments.Segment(2, 3, -1), segments.Segment(0, 1, -1)},
                    {segments.Segment(0, 1, -1), segments.Segment(2, 3, -1)})
    _sort_and_merge({segments.Segment(0, 1, -1), segments.Segment(1, 2, -1)},
                    {segments.Segment(0, 2, -1)})
    _sort_and_merge({segments.Segment(0, 3, -1), segments.Segment(1, 5, -1)},
                    {segments.Segment(0, 5, -1)})


def _find_matching(text, keywords, ignore_case, expected):
    actual = segments.find_matching(text, keywords, ignore_case, -1)
    assert actual[0] == expected[0]
    assert actual[1] == expected[1]


def test_find_matching():
    '''Test segments.find_matching()'''
    no_match = (False, [])
    # Checking matching empty line works
    _find_matching('\n', ['^\n'], False, (True, [(0, 1, -1)]))
    # Checking matching null string does not crash program
    _find_matching('012345', ['^'], False, no_match)
    _find_matching('012345', ['$'], False, no_match)
    # Basic non matching cases
    _find_matching('Some text', {'Not matching'}, False, no_match)
    _find_matching('abcde', {'cdefgh'}, False, no_match)
    _find_matching('abcde', {'0123abc'}, False, no_match)
    # Checking case sensitive matching
    _find_matching('a', ['a'], False, (True, [(0, 1, -1)]))
    _find_matching('a', ['A'], True, (True, [(0, 1, -1)]))
    _find_matching('a', ['A'], False, no_match)
    # Basic matching
    _find_matching('abcde', {'abcde'}, False, (True, [(0, 5, -1)]))
    _find_matching('abcde', {'bcd'}, False, (True, [(1, 4, -1)]))
    _find_matching('abcde', {'a', 'c', 'e'}, False,
                   (True, [(0, 1, -1), (2, 3, -1), (4, 5, -1)]))
    _find_matching('abcde', {'b', 'd'}, False,
                   (True, [(1, 2, -1), (3, 4, -1)]))
    # Overlapping matches
    _find_matching('abcdef', {'bcd', 'cde'}, False,
                   (True, [(1, 5, -1)]))


def make_s(args):
    '''Helper function to make segment'''
    return segments.Segment._make(args)


def test_merge():
    '''Test segments.merge()'''

    merged = segments.merge([], [])
    assert len(merged) == 0

    merged = segments.merge([make_s((1, 2, 3))], [])
    assert len(merged) == 1
    assert merged[0] == (1, 2, 3)

    merged = segments.merge([], [make_s((1, 2, 3))])
    assert len(merged) == 1
    assert merged[0] == (1, 2, 3)

    # bottom clear ahead
    merged = segments.merge([make_s((1, 2, 0))], [make_s((10, 20, 1))])
    assert len(merged) == 2
    assert merged[0] == (1, 2, 0)
    assert merged[1] == (10, 20, 1)

    # top covers whole bottom
    merged = segments.merge([make_s((10, 20, 0))], [make_s((0, 30, 1))])
    assert len(merged) == 1
    assert merged[0] == (0, 30, 1)

    # top covers middle of bottom
    merged = segments.merge([make_s((0, 30, 0))], [make_s((10, 20, 1))])
    assert len(merged) == 3
    assert merged[0] == (0, 10, 0)
    assert merged[1] == (10, 20, 1)
    assert merged[2] == (20, 30, 0)

    # top covers only start of bottom
    merged = segments.merge([make_s((10, 20, 0))], [make_s((5, 15, 1))])
    assert len(merged) == 2
    assert merged[0] == (5, 15, 1)
    assert merged[1] == (15, 20, 0)

    # top covers only end of bottom
    merged = segments.merge([make_s((10, 20, 0))], [make_s((15, 25, 1))])
    assert len(merged) == 2
    assert merged[0] == (10, 15, 0)
    assert merged[1] == (15, 25, 1)

    # multiple tops that only covers middle of bottom
    merged = segments.merge(
        [make_s((10, 30, 0))],
        [make_s((12, 20, 1)), make_s((25, 28, 1))])
    assert len(merged) == 5
    assert merged[0] == (10, 12, 0)
    assert merged[1] == (12, 20, 1)
    assert merged[2] == (20, 25, 0)
    assert merged[3] == (25, 28, 1)
    assert merged[4] == (28, 30, 0)

    # Overlap start
    merged = segments.merge(
        [make_s((10, 20, 0))],
        [make_s((0, 1, 1)), make_s((2, 3, 1))])
    assert len(merged) == 3
    assert merged[0] == (0, 1, 1)
    assert merged[1] == (2, 3, 1)
    assert merged[2] == (10, 20, 0)


def test_flatten():
    '''Test segments.flatten()'''

    merged = segments.flatten(
        [
            [make_s((0, 30, 0))],
            [make_s((5, 25, 1))],
            [make_s((10, 20, 2))]
        ]
    )
    assert len(merged) == 5
    assert merged[0] == (0, 5, 0)
    assert merged[1] == (5, 10, 1)
    assert merged[2] == (10, 20, 2)
    assert merged[3] == (20, 25, 1)
    assert merged[4] == (25, 30, 0)

    merged = segments.flatten(
        [
            [make_s((0, 5, 1))],
            [
                make_s((3, 4, 0)),
                make_s((10, 11, 0))
            ]
        ]
    )
    assert len(merged) == 4, merged
    assert merged[0] == (0, 3, 1)
    assert merged[1] == (3, 4, 0)
    assert merged[2] == (4, 5, 1)
    assert merged[3] == (10, 11, 0)



'''Module holding general utilities'''

import sys


def wrapper(force, func, /, *args):
    '''Helper function that wraps given function (eg curses.wrapper) so
    that print() statements gets bufferized and dumped to stdout in
    case func raises any exception. This allows seeing the output of
    debug print() statements which are otherwise lost as curses traps
    stdout for its own use.'''

    class StdBuffer:
        '''Helper class to bufferize stdout and stderr'''
        def __init__(self):
            self._lines = []

        def write(self, line):
            '''Write current line'''
            self._lines.append(line)

        def get(self):
            '''Get all the lines that were bufferized'''
            return ''.join(self._lines)

    buf = StdBuffer()
    sys.stdout = buf
    sys.stderr = buf

    error = None
    # pylint: disable=broad-except
    try:
        func(*args)
    except Exception as ex:  # pragma: no cover
        error = ex

    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__

    if force or error:
        print('== Start of bufferized stdout and stderr ==')
        print(buf.get())
        print('== End of bufferized stdout and stderr ==')

    if error:  # pragma: no cover
        print('== Exception ==')
        raise error



'''Unit tests for keys'''

import time
import curses

from .. import keys
from .. import enums


class MouseProvider:
    '''Class providing a getmouse() function similar to curses.getmouse().'''
    def getmouse(self):
        '''Mocks curses.getmouse().'''
        assert False  # pragma: no cover


def test_processor():
    '''Test keys.Processor'''
    proc = keys.Processor(keys.Provider([ord('a')]), MouseProvider())
    assert proc
    assert ord('a') == proc.get().key


def test_process():
    '''Test keys.Process'''
    proc = keys.Processor(keys.Provider([]), MouseProvider())
    assert proc

    key = proc.process(-1)
    assert key.key == -1

    key = proc.process(ord('a'))
    assert key.key == ord('a')
    key = proc.process(ord('A'))
    assert key.key == ord('A')

    for k in '\x1b[1;2C':
        key = proc.process(ord(k))
        assert key.key == -1 or key.cmd == enums.Command.GO_SRIGHT

    key = proc.process(ord('A'))
    assert key.key == ord('A')

    for k in '\x1b[1;2C':
        key = proc.process(ord(k))
        assert key.key == -1 or key.cmd == enums.Command.GO_SRIGHT

    for k in '\x1bZ\x1b':
        key = proc.process(ord(k))
        assert key.key == -1

    # Make sure we spit out ESC if polling right after escaping
    key = proc.process(27)
    assert key.key == keys.POLL
    key = proc.process(-1)
    assert key.key == curses.ascii.ESC

    # Make sure we spit out UNMAP if polling in middle of unrecognize seq
    key = proc.process(27)
    assert key.key == keys.POLL
    key = proc.process(ord('['))
    assert key.key == keys.POLL
    key = proc.process(-1)
    assert key.key == keys.UNMAP

    # Make sure we timeout while escaping
    keys.ESCAPE_TIMEOUT = 0.01
    key = proc.process(27)
    assert key.key == -1
    time.sleep(0.05)
    key = proc.process(ord('A'))
    assert key.key == ord('A')



'''Provides processor for Select Graphic Rendition (SGR)

https://en.wikipedia.org/wiki/ANSI_escape_code#SGR
'''
import curses
import re

from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

from . import colors
from . import enums
from . import segments


SGR_COLOR_TO_COLOR_ID: Dict[int, colors.ColorId] = {
    30: 0,
    31: 1,
    32: 2,
    33: 3,
    34: 4,
    35: 5,
    36: 6,
    37: 7,
    39: 0,
    90: 8,
    91: 9,
    92: 10,
    93: 11,
    94: 12,
    95: 13,
    96: 14,
    97: 15,
    40: 16,
    41: 17,
    42: 18,
    43: 19,
    44: 20,
    45: 21,
    46: 22,
    47: 23,
    100: 24,
    101: 25,
    102: 26,
    103: 27,
    104: 28,
    105: 29,
    106: 30,
    107: 31,
}

SGR_BG_COLOR_TO_COLOR_ID: Dict[int, colors.ColorId] = {
    40: 0,
    41: 1,
    42: 2,
    43: 3,
    44: 4,
    45: 5,
    46: 6,
    47: 7,
    100: 8,
    101: 9,
    102: 10,
    103: 11,
    104: 12,
    105: 13,
    106: 14,
    107: 15,
}

FG_BG_TO_COLOR_PAIR_FUNC: \
    Callable[[colors.ColorId, colors.ColorId], colors.Pair] = \
    colors.get_fg_bg_color_pair


class Processor:
    '''Holds original file coloring, built by parsing CSI.'''
    # https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_(Control_Sequence_Introducer)_sequences
    # https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters

    def __init__(self) -> None:
        self._a: int = 0
        self._s: int = -1
        self._fg: int = -1
        self._bg: int = -1

    def filter(self,
               line,
               mode: enums.SgrMode,
               ) -> Tuple[List[segments.Segment], str]:
        '''Process the SGR from the given line and returns the resulting raw
        string and the segments to render the requested attributes.

        '''
        background: List[segments.Segment] = []
        if mode == enums.SgrMode.NONE:
            return background, line
        new_line = re.sub(r'\x1b\[[0-?]*[!-/]*[@-~]', '', line)
        if mode == enums.SgrMode.REMOVE:
            return background, new_line

        assert mode == enums.SgrMode.PROCESS
        eaten = 0

        def seg_end(end):
            start, attr = self._s, self._a
            if 0 <= start < end:
                assert start < end, f'{start} {end}'
                assert attr != 0
                background.append(segments.Segment(start, end, attr))
            self._s = -1

        def set_attr(a):
            if not a:
                return
            a = int(a)
            if a == 0:
                self._a = 0
                self._fg = -1
                self._bg = -1
            elif a == 1:
                self._a = self._a | curses.A_BOLD
            elif a == 2:
                self._a = self._a | curses.A_DIM
            elif a == 4:
                self._a = self._a | curses.A_UNDERLINE
            elif a in SGR_COLOR_TO_COLOR_ID:
                if a in SGR_BG_COLOR_TO_COLOR_ID:
                    self._bg = SGR_BG_COLOR_TO_COLOR_ID[a]
                else:
                    self._fg = SGR_COLOR_TO_COLOR_ID[a]
                pair = FG_BG_TO_COLOR_PAIR_FUNC(self._fg, self._bg)
                self._a = (~curses.A_COLOR & self._a) | pair

        # If we have attribute from previous line, start a pending segment
        if self._a:
            self._s = 0

        for match in re.finditer(r'\x1b\[([0-9]+)(;([0-9]+))?m', line):
            s, e = match.start(), match.end()
            assert s < e
            a_1, a_2 = match.group(1), match.group(3)
            start = s - eaten
            eaten += e - s
            # Finish any pending segment
            seg_end(start)
            # Mark start a new pending
            set_attr(a_1)
            set_attr(a_2)
            if self._a:
                self._s = start

        # Finish any pending segment
        seg_end(len(new_line))
        return background, new_line



'''Provides various model classes for searchf application.

Model classes are used to organize content processing. The main
classes are:
- RawContent: holds the unfiltered original file content.
- SelectedContent: holds all the lines that got selected either because
  they match a filter or because of additional context being revealed.
- DisplayContent: line-wrapped version of SelectedContent.
'''

import math

from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

from . import enums
from . import segments
from . import types
from . import sgr


class DisplayLine(NamedTuple):
    '''Holds data associated with a single displayable line on the screen. Each
    line in the original file might not entirely fit on screen, and can be
    split into several DisplayLine (when line wrapping is on). In other words,
    several DisplayLine might refer to same selected line and have the same
    line_index.

    Attributes:
        line_index  Index of line in array of SelectedLines
        offset      Offset in the original line of the first char of this
                    displayable line.
    '''
    line_index: int
    offset: int


class DisplayContent:
    '''Displayable content. Highest level content related class that is not
    dependent on curses. DisplayContent instances are created by layout
    operations applied to SelectedContent, which basically boils down to
    performing line wrapping.

    For each selected line we need to identify the first display line
    associated with it, and we use firstdlines for this. So there is
    exactly one element in firstdlines for each SelectedLine.

    Attributes:
        firstdlines  First display line of each selected lines.
        dlines       List of display lines.
    '''
    def __init__(self) -> None:
        self.firstdlines: List[int] = []
        self.dlines: List[DisplayLine] = []

    def lines_count(self) -> int:
        '''Gets the number of lines required to display the whole content
        without clipping any of it.'''
        return len(self.dlines)


class Filter:
    '''Filters are used to select lines and highlight keywords in these
    matching lines. In practice, each filter holds properties defining
    how matching is done and a list of keywords. Keywords can be added
    or removed by end-user.

    Attributes:
        ignore_case  Whether or not matching is done ignoring case.
        hiding       Defines the visibility of lines matching this filter.
        keywords     List of keywords.
    '''
    ignore_case: bool = False
    hiding: bool = False

    def __init__(self) -> None:
        # We use a Dict to make sure that keywords are never added
        # twice (ie like a set) and that insertion order is preserved
        # (ie like a stack).
        self.keywords: Dict[str, None] = {}

    def add(self, keyword: str) -> None:
        '''Adds given keyword to this filter.'''
        self.keywords[keyword] = None

    def pop(self) -> Tuple[str, None]:
        '''Removes most recently added keyword from this filter.'''
        return self.keywords.popitem()

    def get_count_and_last_keyword(self) -> Tuple[int, Optional[str]]:
        '''Returns the number of keywords and the last keyword.'''
        count = len(self.keywords)
        if count <= 0:
            return 0, None
        # No peek API, so we need to pop and add back...
        keyword, _ = self.pop()
        self.add(keyword)
        return count, keyword


def digits_count(number: int) -> int:
    '''Returns the number of digits required to display given number.'''
    return math.floor(math.log10(max(1, number))+1)


RULER_INDEX = -1


class SelectedLine(NamedTuple):
    '''Data associated with a selected line. A line can be selected either
    because it directly matches a filter, or because it is in proximity of
    another matching line and user has specify a LineVisibility to reveal
    context.

    Attributes:
        line_index    The index of the line in the original content, or -1
                      if this line does not represent original content (like
                      an horizontal ruler).
        filter_index  The index of the first matching filter if any, or -1
                      otherwise
        text          The actual raw text of the line.
        segments      The segments that will be highlighted.
    '''
    line_index: int
    filter_index: int
    text: str
    segments: List[segments.Segment]


VISIBILITY_TO_SIZE = {
    enums.LineVisibility.ONLY_MATCHING: 0,
    enums.LineVisibility.CONTEXT_1: 1,
    enums.LineVisibility.CONTEXT_2: 2,
    enums.LineVisibility.CONTEXT_5: 5,
    enums.LineVisibility.ALL: -1,
}


class SelectedLineQueue:
    '''Class used to filter out SelectedLine according to a given line
    visibility mode. SelectedLines are added sequentially, one by one,
    and are then yielded back to caller (or not) so as to reveal the
    desired amount of non-matching lines above and below matching
    lines.
    '''
    def __init__(self, mode: enums.LineVisibility):
        self._queue: List[SelectedLine] = []
        assert mode in VISIBILITY_TO_SIZE
        self._size = VISIBILITY_TO_SIZE[mode]
        self._left: int = 0
        self._last_line_visible: int = 0

    def _add(self, selected_line: SelectedLine) -> bool:
        '''Adds the given selected line and returns whether or not
        the caller should flush the queue.'''
        if self._size == 0:
            return False
        self._queue.append(selected_line)
        if self._size < 0:
            # Infinite capacity, never queueing anything...
            return True
        if self._left > 0:
            # Want caller to call flush
            self._left -= 1
            return True
        # We are now queueing these lines, as we are unsure they will be shown
        if len(self._queue) > self._size:
            self._queue.pop(0)
        return False

    def _flush(self) -> List[SelectedLine]:
        lines = self._queue
        self._queue = []
        return lines

    def _update_last_line_visible(self, selected_line: SelectedLine):
        idx, _, _, _ = selected_line
        self._last_line_visible = idx + 1

    def add_matching(self, selected_line: SelectedLine) -> List[SelectedLine]:
        '''Adds a line that matches a filter and returns the list of lines
        that are visible as a result of this operation. At a minimum, the
        returned list of lines only contains the added line. But if more
        context is revealed with LineVisibility setting, more lines can
        be returned.'''
        lines = []
        queued = self._flush()
        if len(queued) > 0:
            # Check if we need to add horizontal rule (index -1)
            idx, _, _, _ = queued[0]
            if idx > self._last_line_visible:
                lines.append(SelectedLine(RULER_INDEX, -1, '', []))
            lines = lines + queued

        lines.append(selected_line)
        self._update_last_line_visible(selected_line)

        # Make sure we are going to bufferize the appropriate amount
        # of subsequent non-matching lines, if we encounter any.
        self._left = self._size

        return lines

    def add_non_matching(self,
                         selected_line: SelectedLine
                         ) -> List[SelectedLine]:
        '''Adds a line that does not match any filter. Returns the list of
        lines that are visible as a result of this operation, if any.'''
        if not self._add(selected_line):
            return []
        queued = self._flush()
        assert len(queued) == 1
        self._update_last_line_visible(selected_line)
        return queued


class SelectedContent:
    '''Holds filtered content of a file, and generates DisplayContent
    instances according a desired screen layout.
    '''

    def reset(self,
              lines: List[SelectedLine],
              hits: List[int]
              ) -> None:
        '''Reset the selected content with given selected lines and hits
        count.

        '''
        self._lines = lines
        self._hits = hits

    @property
    def hits(self) -> List[int]:
        '''How many times each filter was matched'''
        return self._hits

    @property
    def lines(self) -> List[SelectedLine]:
        '''List of SelectedLine that each contains information on what segment
        of the original line matches which filter.'''
        return self._lines

    def __init__(self) -> None:
        self.reset([], [])

    def visible_line_count(self) -> int:
        '''Gets the total number of lines that would be visible if not
        limited by screen height'''
        return len(self._lines)

    def hits_count(self) -> int:
        '''Gets the current hits count'''
        return sum(self._hits)

    def layout(self,
               height: int,
               width: int,
               wrapping: bool
               ) -> DisplayContent:
        '''Generate displayable content, breaking the given selected lines
        into displayable lines taking into account the available space
        on the screen.'''

        assert height > 0
        assert width > 0

        dlines: List[DisplayLine] = []
        firstdlines: List[int] = []
        if not wrapping:
            for idata, mdata in enumerate(self._lines):
                firstdlines.append(len(dlines))
                dlines.append(DisplayLine(idata, 0))
        else:
            for idata, mdata in enumerate(self._lines):
                firstdlines.append(len(dlines))
                line_idx, _, text, _ = mdata
                offset = 0
                if line_idx == RULER_INDEX:
                    dlines.append(DisplayLine(idata, offset))
                    continue
                left = len(text)
                # If the line is empty, we still push a DisplayLine for
                # proper rendering of empty lines
                while True:
                    dlines.append(DisplayLine(idata, offset))
                    offset += width
                    left -= width
                    if left <= 0:
                        break
        assert len(firstdlines) == len(self._lines)
        dc = DisplayContent()
        dc.dlines = dlines
        dc.firstdlines = firstdlines
        return dc


def apply_filters(
        line: str,
        background: List[segments.Segment],
        filters: List[Filter],
        hits: List[int],
) -> Tuple[bool, int, List[segments.Segment]]:
    '''Apply filters to the line and return true if the line should be
    displayed, with the index of first filter that has been match (or -1 if
    none) and the list of segments to colorize the line with. This function
    increments the hit count of each matching filter.

    '''
    first_show_idx = -1
    hide_count = 0
    show_count = 0
    segs = []
    segs.append(background)
    for fidx, f in enumerate(filters):
        matching, matching_segments = \
            segments.find_matching(line, f.keywords, f.ignore_case, fidx)
        if matching:
            hits[fidx] += 1
            if f.hiding:
                hide_count += 1
            else:
                show_count += 1
                segs.append(matching_segments)
                if first_show_idx < 0:
                    first_show_idx = fidx
    if show_count > 0:
        assert first_show_idx >= 0
        assert len(segs) > 0
        return True, first_show_idx, segments.flatten(segs)
    if hide_count <= 0:
        return True, -1, background
    return False, -1, background


class RawContent:
    '''Holds raw content of a file and generates instances of
    SelectedContent from filters.

    Attributes:
        _lines       Lines of the original file.
    '''
    def __init__(self) -> None:
        self._lines: List[str] = []
        self._sgr = sgr.Processor()

    def line_count(self) -> int:
        '''Gets the number of lines in the original file'''
        return len(self._lines)

    def line_number_length(self) -> int:
        '''Number of digit required to display bigest line number'''
        return digits_count(len(self._lines))

    def set_lines(self, lines: List[str]) -> None:
        '''Sets and stores the file content lines.'''
        self._lines = lines

    def filter(self,
               filters: List[Filter],
               line_mode: enums.LineVisibility,
               sgr_mode: enums.SgrMode,
               ) -> SelectedContent:
        '''Filters the raw content using the given filters and processing modes
        (line visibility, SGR) and returns an instance of SelectedContent.

        '''
        lines: List[SelectedLine] = []
        hits = [0 for f in filters]
        line_mode = line_mode if sum(not f.hiding for f in filters) > 0 \
            else enums.LineVisibility.ALL
        line_queue = SelectedLineQueue(line_mode)
        for i, line in enumerate(self._lines):
            # Replace tabs with 4 spaces (not clean!!!)
            line = line.replace('\t', '    ')
            background, line = self._sgr.filter(line, sgr_mode)
            assert len(line) <= 0 or ord(line[0]) != 0, \
                f'Line {i} has embedded null character'
            shown, fidx, segs = apply_filters(line, background, filters, hits)
            if shown:
                if fidx >= 0:
                    lines += line_queue.add_matching(
                        SelectedLine(i, fidx, line, segs))
                else:
                    lines += line_queue.add_non_matching(
                        SelectedLine(i, -1, line, segs))
        assert len(filters) == len(hits)
        sc = SelectedContent()
        sc.reset(lines, hits)
        return sc


class ViewConfig:
    '''This class holds the configuration of a view, like filters to use or the
    display modes, typically changed by end users to match their need. Does not
    contain any data related to actual file content, and can get serialized for
    persistence (see storage.py). This class is not a view class per-se in the
    sense that it does NOT depend on curses or anything UI related.
    '''

    # pylint: disable=too-many-instance-attributes

    line_numbers: bool = True
    wrap: bool = True
    bullets: bool = False
    reverse_matching: bool = False
    line_visibility: enums.LineVisibility = enums.LineVisibility.ONLY_MATCHING
    show_all_lines: bool = True
    show_spaces: bool = False
    sgr_mode: enums.SgrMode = enums.SgrMode.PROCESS
    colorize_mode: enums.ColorizeMode = enums.ColorizeMode.KEYWORD_HIGHLIGHT
    palette_id: types.PaletteId = 0
    dirty: bool = False

    def __init__(self) -> None:
        self.filters: List[Filter] = []

    def get_filters_count(self) -> int:
        '''Returns the number of filters currently defined.'''
        return len(self.filters)

    def has_filters(self) -> bool:
        '''Returns whether or not there is any filter.'''
        return len(self.filters) > 0

    def top_filter(self) -> Filter:
        '''Returns top level filter (most recently added).'''
        count = len(self.filters)
        assert count > 0
        return self.filters[count-1]

    def push_filter(self, f: Filter) -> None:
        '''Pushes the given filter.'''
        self.dirty = True
        self.filters.append(f)

    def swap_filters(self, i, j) -> None:
        '''Swaps the given filters.'''
        self.dirty = True
        count = len(self.filters)
        assert 0 <= i < count
        assert 0 <= j < count
        self.filters[i], self.filters[j] = self.filters[j], self.filters[i]

    def rotate_filters(self, go_up: bool) -> None:
        '''Rotates filters.'''
        self.dirty = True
        count = len(self.filters)
        assert count >= 2
        if go_up:
            self.filters = self.filters[1:] + self.filters[:1]
        else:
            self.filters = self.filters[-1:] + self.filters[:-1]

    def set_palette(self, pid: types.PaletteId) -> None:
        '''Select the next palette in the given direction.'''
        self.dirty = True
        self.palette_id = pid


class Offsets:
    '''Class managing horizontal and vertical offsets in displayable
    content.

    Attributes:
        hoffset      Horizontal offset in content (index of first visible
                     column).
        voffset      Vertical offset in content (index of first visible
                     line).
    '''
    def __init__(self) -> None:
        self.hoffset: int = 0
        self.voffset: int = 0
        self.voffset_desc: str = ''
        self._ymax: int = 0

    def layout(self, available_height: int, content_lines_count: int) -> None:
        '''Layout function defining the space available for the
        displayable content.'''
        assert available_height >= 0
        assert content_lines_count >= 0
        self._ymax = max(0, content_lines_count - available_height)

    def set_h_offset(self, offset: int) -> bool:
        '''Sets the horizontal offset.
        Returns True if changed, False otherwise.'''
        offset = max(offset, 0)
        if self.hoffset == offset:
            return False
        self.hoffset = offset
        return True

    def set_v_offset(self, offset: int) -> bool:
        '''Sets the vertical offfset.
        Returns True if changed, False otherwise.'''
        if offset >= self._ymax:
            offset = self._ymax
            desc = 'BOT'
        elif offset <= 0:
            offset = 0
            desc = 'TOP'
        else:
            percent = int(offset * 100 / self._ymax)
            desc = f'{percent}%'
        if self.voffset == offset:
            return False
        self.voffset = offset
        self.voffset_desc = desc
        return True



'''Unit tests for storage'''

from .. import storage


def test_store():
    '''Test store'''
    store = storage.Store('.searchf.test')
    store.destroy()
    assert not store.can_load()

    # Check we can save and then load
    slot_id = store.save({'text': 'My object 1'})
    assert store.can_load()
    assert slot_id == 0

    # Check we load expected object
    obj, slot_id2 = store.load(False)
    assert slot_id == slot_id2
    assert obj['text'] == 'My object 1'

    # Check we can delete
    slot_id3 = store.delete()
    assert slot_id == slot_id3

    # Check we cannot delete when we have no current slot
    assert not store.delete()

    # Save a new object
    slot_id = store.save({'text': 'My object 2'})
    assert slot_id == 0

    # Create a brand new store and check we can load
    store = storage.Store('.searchf.test')
    assert store.can_load()
    obj, slot_id = store.load(False)
    assert slot_id == 0
    assert obj['text'] == 'My object 2'

    # Check destroying
    store.destroy()



'''Application end-to-end tests.
'''

import argparse
import curses
import os
import time

from typing import List
from typing import NamedTuple
from contextlib import contextmanager

from .. import app
from .. import colors
from .. import debug
from .. import keys
from .. import main
from .. import storage
from .. import types
from .. import utils

from . import test_enums
from . import test_keys
from . import test_models
from . import test_segments
from . import test_sgr
from . import test_storage

TEST_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                         'sample.txt')
TEST_FILE_C = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                           'sample_with_colors.txt')


class KeywordsInjector:
    '''Class that emulates end-user entering keywords. Provides a
    get_text() function that can replace the one in the main module.
    '''
    def __init__(self, keywords: List[str]):
        self._keywords = keywords

    def get_next(self) -> str:
        '''Returns the next keyword'''
        return self._keywords.pop(0)

    def get_text(self, scr, y, x, text_prompt, handler, text) -> str:
        '''Function to replace main.get_next'''
        # pylint: disable=unused-argument
        return self.get_next()


class MtimeInjector:
    '''Class that emulates end-user externally changing file. Provides
    a getmtime() function that can replace the one in the main module,
    and that makes sure the file always changes.'''

    def __init__(self):
        self._time = 0.0

    def getmtime(self, path: str) -> float:
        '''Returns the file last modification time'''
        # pylint: disable=unused-argument
        self._time += 1
        return self._time


class MouseProvider:
    '''Class providing a getmouse() function similar to curses.getmouse()'''
    def __init__(self, states: List[int]) -> None:
        self._states: List[int] = states

    def getmouse(self):
        '''Mocks curses.getmouse().'''
        return None, 0, 0, 0, self._states.pop(0)


# Screen size used for tests, terminal must be bigger for
# tests to run
TEST_SIZE = (30, 60)


@contextmanager
def main_modifier(stdscr,
                  keywords_injector: KeywordsInjector,
                  mtime_injector: MtimeInjector):
    '''Context manager that replaces key and text input methods of the main
    module, and makes sure tests are run always with same screen
    resolution.
    '''

    size = app.get_max_yx(stdscr)
    assert size[0] >= TEST_SIZE[0], f'{size} < {TEST_SIZE}'
    assert size[1] >= TEST_SIZE[1], f'{size} < {TEST_SIZE}'

    def _injected_get_max_yx(_):
        return TEST_SIZE

    get_text = app.get_text
    getmtime = app.getmtime
    get_max_yx = app.get_max_yx
    app.get_text = keywords_injector.get_text
    app.getmtime = mtime_injector.getmtime
    app.get_max_yx = _injected_get_max_yx
    try:
        yield
    finally:
        app.get_text = get_text
        app.getmtime = getmtime
        app.get_max_yx = get_max_yx


class AppTest(NamedTuple):
    '''Model data associated with a test:
    - the description of the test
    - the list of keys that will be sequentially automatically pressed
      by test runner
    - the list of text input that will be automatically fed to the
      application
    '''
    description: str
    keys: List[str]
    inputs: List[str]


HANDLE_EVENT_DELAY = 0.01


def _run(stdscr, t: AppTest, f: str):
    '''Helper function to run the given AppTest'''
    print(t.description)
    stdscr.clear()
    colors.init()
    store = storage.Store('.searchf.test')
    with main_modifier(stdscr,
                       KeywordsInjector(t.inputs.copy()),
                       MtimeInjector()):
        margins = types.Margins()
        margins.bottom += 1
        main.APP.create(store, stdscr, margins, True, f)
        for key in t.keys:
            stdscr.refresh()
            # Add sleep just to see something, test can run without it
            time.sleep(HANDLE_EVENT_DELAY)
            i = key if isinstance(key, int) else ord(key)
            main.APP.handle_event(keys.KeyEvent(i))
    store.destroy()


def _test_main_init_env():
    print('Test main.init_env()')
    parser = main.init_env()
    assert parser


def _test_main_get_text(stdscr):
    print('Test main.get_text()')

    def my_handler(_):
        pass

    def my_handler_throwing(_):
        raise app.EscapeException

    app.get_text(stdscr, 0, 0,
                 "Testing prompt", my_handler, 'Editable content')
    app.get_text(stdscr, 0, 0,
                 "Testing prompt", my_handler_throwing, '')


def _test_app_validate():
    print('Test app.validate()')
    assert app.validate('a') == 'a'
    assert app.validate(curses.ascii.DEL) == curses.KEY_BACKSPACE
    actual = None
    try:
        app.validate(curses.ascii.ESC)
    except app.EscapeException as ex:
        actual = ex
    assert actual


def _test_main_main_loop(stdscr):
    print('Test main.main_loop()')
    keys_processor = keys.Processor(keys.Provider(
        [' ', '>', 'l', keys.POLL, 'q']),
                                    MouseProvider([]))
    main.main_loop(stdscr, TEST_FILE, False, False, keys_processor)


def _test_main_resize(stdscr):
    print('Test main.resize')
    keys_processor = keys.Processor(keys.Provider([curses.KEY_RESIZE, 'q']),
                                    MouseProvider([]))
    main.main_loop(stdscr, TEST_FILE, False, False, keys_processor)


def _test_main_mouse(stdscr):
    print('Test main.mouse')
    keys_processor = keys.Processor(keys.Provider(
        [curses.KEY_MOUSE, curses.KEY_MOUSE, 'q']),
                                    MouseProvider(
        [curses.BUTTON4_PRESSED, 123456]))
    main.main_loop(stdscr, TEST_FILE, False, False, keys_processor)


def _test_main_debug_view(stdscr):
    # Test debug mode in a very hacky way by hijacking handle_event function
    # and spitting out a few dummy debug lines per key press
    app.USE_DEBUG = True
    original_handle_event = main.APP.handle_event

    def my_handle_event(key):
        for i in range(20):
            debug.out(f'Test {i} dbg {key}')
        return original_handle_event(key)

    main.APP.handle_event = my_handle_event
    app_test = AppTest(
        'Test special debug mode',
        ['/', 'n', 'n', 'n', 'p', 'p'], ['filter'])
    _run(stdscr, app_test, TEST_FILE)
    main.APP.handle_event = original_handle_event
    app.USE_DEBUG = False


# This is poor man's testing, as we don't validate much other than
# just making sure things don't blow up when executing common commands
# (cf all assert in code). By maintaining code coverage above 95%,
# these tests are still very useful to catch regression when
# refactoring.
def _run_app_tests(stdscr):

    # Log param, that might be useful to repro any issue
    print(f'TEST_FILE = {TEST_FILE}')
    print(f'stdscr.getmaxyx() = {stdscr.getmaxyx()}')
    print()

    app_tests = [
        AppTest('Test keywords that are invalid regex',
                ['f'], ['?']),
        AppTest('Test that help can get displayed',
                ['?', 'q'], []),
        AppTest('Test view switching',
                ['r', 't', '1', '2', '3', '!', '@', '#'], []),
        AppTest('Test reloading',
                ['r', 't', keys.POLL, 'R', keys.POLL, 'R', 'T', 'T'], []),
        AppTest('Test scrolling around',
                ['>', '<', ' ', 'b'], []),
        AppTest('Test scrolling horizontally',
                [curses.KEY_RIGHT, curses.KEY_LEFT], []),
        AppTest('Test goto lines',
                ['\t', '\t', '\t', '\t'], ['5', '99999', 'bad', '0']),
        AppTest('Test various display modes',
                ['l', 'k', 'k', '.', '.', '*', '*', 'm', 'M', 'm'], []),
        AppTest('Test entering one letter keywords',
                ['+', '+', 'f', 'f', 'v', 'h', 'h', 'v', 'v', 'V', 'c', 'c',
                 'F', '-', '-', '-', '+'],
                ['a', 'b', 'c', 'd', '']),
        AppTest('Test entering keywords',
                ['+', '+', 'f', 'f', 'm', 'h', 'h', 'H', 'm', 'c', 'C', 'F',
                 '-', '-', '-', '+'],
                ['filter', 'keyword', 'for', 'python', '']),
        AppTest('Test entering empty keywords, poping non existent keywords',
                ['+', 'f', '-', 'F'],
                ['', '']),
        AppTest('Test editing keywords',
                ['e', '+', 'e', 'e'],
                ['something', 'good', '']),
        AppTest('Test keyword search',
                ['/', 'n', 'n', 'n', 'p', 'p', 'p', 'p', '/'],
                ['filter']),
        AppTest('Test case sensitive search',
                ['i', '/', 'i', 'i'],
                ['python']),
        AppTest('Test toggling hiden/shown filter',
                ['x', 'f', 'x', 'x'],
                ['python']),
        AppTest('Test swapping filters',
                ['d', 'f', 'f', 'd'],
                ['key', 'python']),
        AppTest('Test rotating filters',
                ['w', 'f', 'f', 'w', 's'],
                ['key', 'python']),
        AppTest('Test save/load/delete slots',
                ['[', '\\', '|', 'f', 'f', '\\', '\\', '[', ']', '|'],
                ['key', 'python']),
        AppTest('Test SGR Processing modes',
                ['`', '`', '`', '~', '~', '~'],
                ['', '']),
    ]

    for test in app_tests:
        _run(stdscr, test, TEST_FILE)
    for test in app_tests:
        _run(stdscr, test, TEST_FILE_C)

    _test_main_init_env()
    _test_main_get_text(stdscr)
    _test_app_validate()
    _test_main_main_loop(stdscr)
    _test_main_resize(stdscr)
    _test_main_mouse(stdscr)
    _test_main_debug_view(stdscr)


def _run_unit_tests():
    print('Test enums.test_get_next_prev()')
    test_enums.test_get_next_prev()
    print('Test enums.test_from_int()')
    test_enums.test_from_int()
    print('Test enums.test_repr()')
    test_enums.test_repr()
    print('Test segments.iterate()')
    test_segments.test_iterate()
    print('Test segments._sort_and_merge()')
    test_segments.test_sort_and_merge()
    print('Test segments.find_matching()')
    test_segments.test_find_matching()
    print('Test segments.test_merge()')
    test_segments.test_flatten()
    print('Test segments.test_flatten()')
    test_segments.test_merge()
    print('Test models.test_filter()')
    test_models.test_filter()
    print('Test models.test_digit_count()')
    test_models.test_digit_count()
    print('Test models.test_model()')
    test_models.test_model()
    print('Test models.test_display_content()')
    test_models.test_display_content()
    print('Test models.test_offsets()')
    test_models.test_offsets()
    print('Test models.test_view_config()')
    test_models.test_view_config()
    print('Test keys.test_processor()')
    test_keys.test_process()
    print('Test keys.test_process()')
    test_keys.test_processor()
    print('Test storage.test_strore()')
    test_storage.test_store()
    print('Test sgr.test_processor()')
    test_sgr.test_processor()


def _test_main_main():
    print('Test main.main()')

    # We use a context manager that replaces main.init_env and
    # keys.Processor in order to be able to load main.main as if it was
    # invoked by end-user

    def _init_env():
        parser = argparse.ArgumentParser(description='Dummy parser')
        parser.add_argument('-file', default=TEST_FILE)
        parser.add_argument('-debug', default=False)
        parser.add_argument('-show-events', default=False)
        print('test_env')
        return parser

    class _KeysProcessor:
        def __init__(self, _getch, _getmouse):
            self._keys = [-1, ord('<')]

        def get(self) -> keys.KeyEvent:
            '''Get next key event from processor'''
            if len(self._keys) <= 0:
                raise KeyboardInterrupt
            return keys.KeyEvent(self._keys.pop(0))

    @contextmanager
    def _main_modifier():
        keys_processor = keys.Processor
        init_env = main.init_env
        keys.Processor = _KeysProcessor
        main.init_env = _init_env
        try:
            yield
        finally:
            keys.Processor = keys_processor
            main.init_env = init_env

    with _main_modifier():
        main.main()


def test_main():
    '''Test entry point'''
    print('== Tests started ==')
    _run_unit_tests()
    utils.wrapper(True, curses.wrapper, _run_app_tests)
    _test_main_main()
    print('== Tests passed ==')


if __name__ == '__main__':
    test_main()



'''Module implementing views for searchf application.'''

import curses
import curses.ascii
import os
import re
import sys

from typing import Any
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

from . import colors
from . import enums
from . import keys
from . import models
from . import segments
from . import types

# Do not use curses.A_BOLD on Windows as it just renders horribly when
# highlighting
USE_BOLD = 0 if sys.platform == 'win32' else curses.A_BOLD

STATUS_EMPTY = ''


def bool_to_text(value: bool) -> str:
    '''Converts a boolean value to text.'''
    return 'enabled' if value else 'disabled'


CASE_COL_TEXTS = ['Case', 'ignored', 'sensitive']
CASE_COL_LEN = len(max(CASE_COL_TEXTS, key=len))

SHOWN_COL_TEXTS = ['Shown', 'no', 'yes']
SHOWN_COL_LEN = len(max(SHOWN_COL_TEXTS, key=len))


class PrefixInfo(NamedTuple):
    '''Line prefix infos.

    Attributes:
        length       Total length of prefix.
        digit_count  Number of digits of the largest line number.
        separator    Separator between things on left (line number,
                     bullet), and things on right (line).
    '''
    length: int
    digit_count: int
    separator: str


class TextView:
    '''Display selected content of a file, using filters and keyword to
    highlight specific text.'''

    # pylint: disable=too-many-instance-attributes

    def __init__(self, store, scr, name: str, path: str) -> None:
        self._raw: models.RawContent = \
            models.RawContent()
        self._selected: models.SelectedContent = \
            models.SelectedContent()
        self._display: models.DisplayContent = \
            models.DisplayContent()
        self._offsets: models.Offsets = models.Offsets()
        self._config: models.ViewConfig = models.ViewConfig()
        self._store = store
        self._scr = scr
        self._name: str = name
        self._basename: str = os.path.basename(path)
        # self._win: Optional[curses._CursesWindow]
        self._win: Any = None
        self._size: types.Size = (0, 0)
        self._content_available_size: types.Size = (0, 0)
        self._ruler: str = ''

    def get_config(self) -> models.ViewConfig:
        '''Gets the config.'''
        return self._config

    def _slot_save(self) -> types.Status:
        '''Save the config.'''
        if not self._config.has_filters():
            return 'No filters to save'
        if not self._config.dirty:
            return 'No change to save'
        idx = self._store.save(self._config)
        self._config.dirty = False
        return f'Filters saved in slot {idx}'

    def _slot_delete(self) -> types.Status:
        idx = self._store.delete()
        if idx is not None:
            assert idx >= 0
            self._config.dirty = True  # Allows saving again
            return f'Slot {idx} deleted'
        return 'No slot loaded. Cannot delete current slot.'

    def _slot_load(self, goto_next: bool) -> types.Status:
        if not self._store.can_load():
            return 'Nothing to load. All slots are empty.'
        self._config, idx = self._store.load(goto_next)
        self._sync(True)
        self._config.dirty = False
        return f'Slot {idx} loaded'

    def name(self) -> str:
        '''Gets the name of the view.'''
        return self._name

    def layout(self, h: int, w: int, y: int, x: int) -> None:
        '''Sets dimension of the view and recompute underlying view model
        without redrawing anything.'''

        # _size stores number of lines and columns available to draw content
        self._size = (h, w)

        if not self._win:
            self._win = curses.newwin(h, w, y, x)
        else:
            self._win.resize(h, w)
            self._layout(False)

    def _get_prefix_info(self) -> PrefixInfo:
        if self._config.line_numbers:
            number_length = self._raw.line_number_length()
            sep = ' │ '
        elif self._config.wrap:
            number_length = 0
            sep = '◆ ' if self._config.bullets else ''  # curses.ACS_DIAMOND
        else:
            number_length = 0
            sep = ''
        return PrefixInfo(number_length + len(sep), number_length, sep)

    def _get_color_pair(self, i: int) -> colors.Pair:
        return colors.get_color_pair(self._config.palette_id, i)

    def _draw_bar(self, y: int) -> None:
        '''Draws the status bar and the filter stack underneath.'''

        _, w = self._size
        style = curses.color_pair(colors.BAR_COLOR_PAIR_ID)
        self._win.hline(y, 0, curses.ACS_HLINE, w, style)

        if not self._config.has_filters():
            self._win.addstr(' No filter ', style)
        else:
            self._win.addstr(f'{self._selected.hits_count():>8}',
                             style)
            self._win.addstr(f' | {"Case":^{CASE_COL_LEN}}', style)
            self._win.addstr(f' | {"Shown":^{SHOWN_COL_LEN}}', style)
            self._win.addstr(' | Keywords ', style)

        # Print from right to left
        def move_left_for(x, text):
            return max(0, x - len(text) - 1)

        x = w
        text = f' {self._name} '
        x = move_left_for(x, text)
        self._win.addstr(y, x, text, style)

        # voffest_desc is not always shown, but at most 5 char long
        text = ''.ljust(5)
        x = move_left_for(x, text)
        if len(self._offsets.voffset_desc) > 0:
            text = f' {self._offsets.voffset_desc:>3} '
            self._win.addstr(y, x, text, style)

        text = f' {self._raw.line_count()} lines '
        x = move_left_for(x, text)
        self._win.addstr(y, x, text, style)

        text = f' {self._basename} '
        x = move_left_for(x, text)
        self._win.addstr(y, x, text, style | USE_BOLD)

        assert len(self._selected.hits) == \
            self._config.get_filters_count()
        for i, f in enumerate(self._config.filters):
            x = 0
            y += 1
            self._win.addstr(y, 0, f'{self._selected.hits[i]:>8}')
            text = CASE_COL_TEXTS[1 if f.ignore_case else 2]
            self._win.addstr(f' | {text:^{CASE_COL_LEN}}')
            text = SHOWN_COL_TEXTS[1 if f.hiding else 2]
            self._win.addstr(f' | {text:^{SHOWN_COL_LEN}} | ')
            text = ' AND '.join(f.keywords)
            color = 0 if f.hiding else self._get_color_pair(i)
            self._win.addstr(text, color)

    def _draw_prefix(
            self,
            y: int,
            prefix_info: PrefixInfo,
            line_idx: int,
            color: int,
            is_first_line: bool,
    ) -> None:
        _, w_index, sep = prefix_info
        if is_first_line and w_index > 0:
            assert line_idx >= 0
            self._win.addstr(y, 0, f'{line_idx:>{w_index}}', color | USE_BOLD)
        self._win.addstr(y, w_index, f'{sep}')

    def _draw_content(
            self,
            pos: types.Position,
            text: str,
            segs: List[segments.Segment],
            offset: int,
            ffidx: int,
            ffcolor: int,
    ) -> None:
        _, width = self._content_available_size
        vend = offset + width
        if self._config.show_spaces:
            text = text.replace(' ', '·')  # Note: this is curses.ACS_BULLET

        x = pos.x
        for match, start, end, attr in segments.iterate(
                offset, vend, segs):
            assert match or attr == -1
            assert start < end
            length = end - start
            if ffidx != -1 and \
               self._config.colorize_mode == enums.ColorizeMode.LINE:
                attr = ffcolor
            elif attr == -1 or (attr & curses.A_ATTRIBUTES) == 0:
                # Assume attr is a filter index
                attr = self._get_color_pair(attr)
            # otherwise use attributes from segment as is
            self._win.addnstr(pos.y, x, text[start:end], length, attr)
            x += length

    def draw(self) -> None:
        '''Draws the view.'''
        # debug.out(f'{self._name} draw {self._offsets.voffset}')
        self._win.clear()

        prefix_info = self._get_prefix_info()
        ymax, _ = self._content_available_size

        # Only draw what we need: we are going to pull at most
        # _content_available_size[1] lines from DisplayContent, possibly
        # less if some content needs multiple lines (wrapping)

        # We iterate over all the display lines using idline
        idline = self._offsets.voffset
        idlinemax = self._display.lines_count()

        for y in range(ymax):
            if idline >= idlinemax:
                break
            iline, offset = self._display.dlines[idline]
            idline += 1

            # iline is the index of the selected line associated with the
            # current display line. offset is the horizontal offset in
            # original content line where this display line starts.  If offset
            # is not 0, this means the original content line is effectively
            # being wrapped on to multiple lines on the screen.

            line_idx, filter_idx, text, segs = \
                self._selected.lines[iline]
            if line_idx == models.RULER_INDEX:
                self._win.addstr(y, 0, self._ruler)
                continue

            # filter_idx is the index of the first filter that is matching
            # current line
            ffcolor = self._get_color_pair(filter_idx)
            self._draw_prefix(y, prefix_info, line_idx, ffcolor, offset == 0)
            offset += self._offsets.hoffset
            self._draw_content(
                types.Position(prefix_info.length, y),
                text,
                segs,
                offset,
                filter_idx,
                ffcolor)

        self._draw_bar(self._content_available_size[0])
        self._win.refresh()

    def _layout(self, redraw: bool) -> None:
        '''Propagate layout changes: evaluates the available space and
        recompute displayable content from the selected lines.
        '''

        self._ruler = '-' * self._size[1]

        # Compute space available for file content
        h, w = self._size
        prefix_len, _, _ = self._get_prefix_info()
        h = max(0, h - 1 - self._config.get_filters_count())
        w = max(0, w - prefix_len)
        self._content_available_size = (h, w)

        self._display = self._selected.layout(h, w, self._config.wrap)
        self._offsets.layout(h, self._display.lines_count())

        # Try preserve current voffset, but make sure we comply with
        # constraints which might just have changed.
        self.set_v_offset(self._offsets.voffset, False)

        assert self._offsets.voffset == 0 or \
            self._offsets.voffset < self._display.lines_count()

        if redraw:
            self.draw()

    def _sync(self, redraw: bool) -> None:
        self._selected = self._raw.filter(
            self._config.filters,
            self._config.line_visibility,
            self._config.sgr_mode)
        self._layout(redraw)

    def set_config(self, config: models.ViewConfig) -> None:
        '''Sets the configuration.'''
        self._config = config
        self._sync(False)

    def set_lines(self, lines: List[str]) -> None:
        '''Sets the content of this view. Assumes the view is offscreen and
        does not trigger a redraw.
        '''
        self._raw.set_lines(lines)
        self._sync(False)

    def _pop_filter(self) -> types.Status:
        if self._config.has_filters():
            self._config.filters.pop()
            self._sync(True)
            return 'Filter removed'
        return 'No filter to remove'

    def push_keyword(self, keyword: str, new_filter: bool) -> types.Status:
        '''Pushes a new keyword in current top level filter (if new_filter is
        False) or in a brand new filter (if new_filter is True).'''
        if len(keyword) <= 0:
            return 'No keyword added'
        try:
            re.compile(keyword)
        except re.error:
            return 'Invalid python regex pattern'
        if not self._config.has_filters():
            new_filter = True  # Force a new filter since we have none
        if new_filter:
            f = models.Filter()
            f.ignore_case = keyword.lower() == keyword
            self._config.push_filter(f)

        self._config.top_filter().add(keyword)
        self._sync(True)
        return 'New filter created' if new_filter else 'Keyword added'

    def _pop_keyword(self) -> types.Status:
        if not self._config.has_filters():
            return 'No keyword to remove'
        f = self._config.top_filter()
        k_count = len(f.keywords)
        if k_count <= 1:
            return self._pop_filter()
        f.pop()
        self._sync(True)
        return 'Keyword removed'

    def get_last_keyword(self) -> Tuple[int, Optional[str]]:
        '''Returns total number of keywords in top level filter (0 if none),
        and the keyword that was last entered by user.
        '''
        if not self._config.has_filters():
            return 0, ''
        f = self._config.top_filter()
        return f.get_count_and_last_keyword()

    def _toggle_line_numbers(self) -> types.Status:
        self._config.line_numbers = not self._config.line_numbers
        self._layout(True)
        return f'Line numbers {bool_to_text(self._config.line_numbers)}'

    def _toggle_wrap(self) -> types.Status:
        self._config.wrap = not self._config.wrap
        self._layout(True)
        return f'Line wrapping {bool_to_text(self._config.wrap)}'

    def _toggle_bullets(self) -> types.Status:
        self._config.bullets = not self._config.bullets
        self._layout(True)
        return f'Bullets {bool_to_text(self._config.bullets)}'

    def _toggle_show_spaces(self) -> types.Status:
        self._config.show_spaces = not self._config.show_spaces
        self.draw()
        return f'Show spaces {bool_to_text(self._config.show_spaces)}'

    def _cycle_colorize_mode(self, forward: bool) -> types.Status:
        f = enums.ColorizeMode.get_next if forward \
            else enums.ColorizeMode.get_prev
        self._config.colorize_mode = f(self._config.colorize_mode)
        self.show()
        return f'Colorize mode: {self._config.colorize_mode}'

    def _cycle_line_visibility(self, forward: bool) -> types.Status:
        f = enums.LineVisibility.get_next if forward \
            else enums.LineVisibility.get_prev
        self._config.line_visibility = f(self._config.line_visibility)
        self._sync(True)
        return f'{self._config.line_visibility}'

    def _cycle_sgr_mode(self, forward: bool) -> types.Status:
        f = enums.SgrMode.get_next if forward \
            else enums.SgrMode.get_prev
        self._config.sgr_mode = f(self._config.sgr_mode)
        self._sync(True)
        return f'{self._config.sgr_mode}'

    def _toggle_ignore_case(self) -> types.Status:
        if not self._config.has_filters():
            return 'Cannot change case sentitivity (no keyword)'
        f = self._config.top_filter()
        f.ignore_case = not f.ignore_case
        self._sync(True)
        return f'Ignore case set to {f.ignore_case}'

    def _toggle_hiding(self) -> types.Status:
        if not self._config.has_filters():
            return 'Cannot change filter property (no filter)'
        f = self._config.top_filter()
        f.hiding = not f.hiding
        self._sync(True)
        action = 'hidden' if f.hiding else 'shown'
        return f'Lines matching filter are now {action}'

    def _apply_palette_and_draw(self) -> None:
        colors.apply_palette(
            self._config.palette_id,
            self._config.colorize_mode == enums.ColorizeMode.KEYWORD_HIGHLIGHT)
        self.draw()

    def show(self) -> None:
        '''Shows the view, after it was hidden by another one'''
        self._apply_palette_and_draw()

    def _cycle_palette(self, forward: bool) -> types.Status:
        '''Select the next palette in the given direction'''
        pid = colors.cycle_palette(self._config.palette_id, forward)
        self._config.set_palette(pid)
        self._apply_palette_and_draw()
        return f'Using color palette #{pid}'

    def _set_h_offset(self, offset: int) -> types.Status:
        if self._offsets.set_h_offset(offset):
            self.draw()
        return STATUS_EMPTY

    def set_v_offset(self, offset: int, redraw: int) -> types.Status:
        '''Sets the vertical offset of the view.'''
        if self._offsets.set_v_offset(offset) and redraw:
            self.draw()
        return STATUS_EMPTY

    def _hscroll(self, delta: int) -> types.Status:
        self._set_h_offset(self._offsets.hoffset + delta)
        return STATUS_EMPTY

    def _vscroll(self, delta: int) -> types.Status:
        self.set_v_offset(self._offsets.voffset + delta, True)
        return STATUS_EMPTY

    def goto_line(self, line_text: str) -> types.Status:
        '''Makes sure the given line is visible, making it the first line on
        the screen unless that would make last line of the file not
        displayed at the bottom.'''
        try:
            line = int(line_text)
        except ValueError:
            return 'Not a number'
        if line <= 0:
            self.set_v_offset(0, True)
        elif line >= self._display.lines_count():
            self.set_v_offset(sys.maxsize, True)
        else:
            voffset = 0
            for iline, _ in self._display.dlines:
                i, _, _, _ = self._selected.lines[iline]
                if i >= line:
                    self.set_v_offset(voffset, True)
                    break
                voffset += 1
        return f'Goto line {line}'

    def has_filters(self) -> bool:
        '''Returns whether the view has any filter or not.'''
        return self._config.has_filters()

    def swap_filters(self) -> types.Status:
        '''Swaps the top 2 filters.'''
        count = self._config.get_filters_count()
        if self._config.get_filters_count() < 2:
            return 'Not enough filters'
        self._config.swap_filters(count - 1, count - 2)
        self._sync(True)
        return 'Filters swapped'

    def rotate_filters(self, go_up: bool) -> types.Status:
        '''Rotate the filters.'''
        if self._config.get_filters_count() < 2:
            return 'Not enough filters'
        self._config.rotate_filters(go_up)
        self._sync(True)
        return 'Filters rotated'

    def _vscroll_to_match(
            self,
            starting: bool,
            direction: int,
    ) -> types.Status:
        idline = self._offsets.voffset
        ilinemax = len(self._selected.lines)
        iline, _ = self._display.dlines[idline]
        if not starting:
            iline += direction
        while 0 <= iline < ilinemax:
            _, fidx, _, _ = self._selected.lines[iline]
            if fidx >= 0:
                idline = self._display.firstdlines[iline]
                break
            iline += direction
        # Remember current offset to be able to return a side-effect
        # description
        prev_offset = self._offsets.voffset
        self.set_v_offset(idline, True)
        if prev_offset == self._offsets.voffset:
            return '(END)' if direction > 0 else '(BEGIN)'
        return 'Next match' if direction > 0 else 'Previous match'

    def search(self, keyword: str) -> types.Status:
        '''Searches for the given keyword. This function assumes the view
        currently has no filter.'''
        assert not self.has_filters()
        if len(keyword) > 0:
            self._config.reverse_matching = False
            self._config.line_visibility = enums.LineVisibility.ALL
            self.push_keyword(keyword, True)
            self._vscroll_to_match(True, 1)
        return STATUS_EMPTY

    def _vpagescroll(self, delta: int) -> types.Status:
        self._vscroll(delta * self._content_available_size[0])
        return STATUS_EMPTY

    def execute(self, command: enums.Command) -> types.Status:
        '''Executes the given command.'''
        dispatch = {
            enums.Command.GO_UP:
                lambda: self._vscroll(-1),
            enums.Command.GO_DOWN:
                lambda: self._vscroll(1),
            enums.Command.GO_LEFT:
                lambda: self._hscroll(-1),
            enums.Command.GO_RIGHT:
                lambda: self._hscroll(1),
            enums.Command.GO_HOME:
                lambda: self.set_v_offset(0, True),
            enums.Command.GO_END:
                lambda: self.set_v_offset(sys.maxsize, True),
            enums.Command.GO_NPAGE:
                lambda: self._vpagescroll(1),
            enums.Command.GO_PPAGE:
                lambda: self._vpagescroll(-1),
            enums.Command.GO_SLEFT:
                lambda: self._hscroll(-20),
            enums.Command.GO_SRIGHT:
                lambda: self._hscroll(20),
            enums.Command.VSCROLL_TO_NEXT_MATCH:
                lambda: self._vscroll_to_match(False, 1),
            enums.Command.VSCROLL_TO_PREV_MATCH:
                lambda: self._vscroll_to_match(False, -1),
            enums.Command.NEXT_COLORIZE_MODE:
                lambda: self._cycle_colorize_mode(True),
            enums.Command.PREV_COLORIZE_MODE:
                lambda: self._cycle_colorize_mode(False),
            enums.Command.NEXT_PALETTE:
                lambda: self._cycle_palette(True),
            enums.Command.PREV_PALETTE:
                lambda: self._cycle_palette(False),
            enums.Command.NEXT_LINE_VISIBILITY:
                lambda: self._cycle_line_visibility(True),
            enums.Command.PREV_LINE_VISIBILITY:
                lambda: self._cycle_line_visibility(False),
            enums.Command.NEXT_SGR_MODE:
                lambda: self._cycle_sgr_mode(True),
            enums.Command.PREV_SGR_MODE:
                lambda: self._cycle_sgr_mode(False),
            enums.Command.POP_FILTER:
                self._pop_filter,
            enums.Command.POP_KEYWORD:
                self._pop_keyword,
            enums.Command.TOGGLE_LINE_NUMBERS:
                self._toggle_line_numbers,
            enums.Command.TOGGLE_WRAP:
                self._toggle_wrap,
            enums.Command.TOGGLE_BULLETS:
                self._toggle_bullets,
            enums.Command.TOGGLE_SHOW_SPACES:
                self._toggle_show_spaces,
            enums.Command.TOGGLE_IGNORE_CASE:
                self._toggle_ignore_case,
            enums.Command.TOGGLE_HIDING:
                self._toggle_hiding,
            enums.Command.SWAP_FILTERS:
                self.swap_filters,
            enums.Command.ROTATE_FILTERS_UP:
                lambda: self.rotate_filters(True),
            enums.Command.ROTATE_FILTERS_DOWN:
                lambda: self.rotate_filters(False),
            enums.Command.SLOT_SAVE:
                self._slot_save,
            enums.Command.SLOT_DELETE:
                self._slot_delete,
            enums.Command.SLOT_LOAD_NEXT:
                lambda: self._slot_load(True),
            enums.Command.SLOT_LOAD_PREV:
                lambda: self._slot_load(False),
        }
        assert command in dispatch, f'command {command}'
        return dispatch[command]()


class DebugView:
    '''Displays few debug lines, convenient to debug layout while curses
    running.'''
    def __init__(self, scr) -> None:
        self._scr = scr
        self._lines: List[str] = []
        self._size: types.Size = (0, 0)
        self._position = (0, 0)
        self._win = curses.newwin(0, 0, 0, 0)
        # self._f = open('.searchf/debug.txt', 'w+')

    def layout(self, size: types.Size, position: Tuple[int, int]) -> None:
        '''Layout the debug view'''
        self._size = size
        self._position = position
        self._win.resize(size[0], size[1])
        self._win.mvwin(position[0], position[1])

    def draw(self) -> None:
        '''Draws the debug view on the screen'''
        _, w = self._size
        for i, line in enumerate(self._lines):
            try:
                self._win.addstr(i, 0, f'{line:<{w}}')
            except curses.error:
                pass

        self._win.refresh()

    def out(self, *argv) -> None:
        '''Outputs the given arg to the debug view.'''
        assert self._scr
        h, _ = self._size
        self._lines.append(*argv)
        # self._f.write(*argv)
        # self._f.write('\n')
        # self._f.flush()
        if len(self._lines) > h:
            self._lines.pop(0)
        self.draw()


class KeyEventView:
    '''Display KeyEvent pressed by end-user.'''
    def __init__(self, scr):
        self._scr = scr
        self._x = 0
        self._y = 0
        self._w = 0

    def layout(self, y: int, x: int, _: int, w: int) -> None:
        '''Layout view.'''
        self._x = x
        self._y = y
        self._w = w

    def show(self, ev: keys.KeyEvent) -> None:
        '''Show key pressed'''
        if not ev.is_poll():
            text = f'    KEY {ev.text:15} {ev.cmd}'
            self._scr.addstr(self._y,
                             self._x,
                             f'{text:<{self._w}}',
                             curses.A_BOLD)



'''Unit tests for enums'''

from .. import enums


class DummyEnum(enums.AutoEnum):
    '''Test enum.'''
    FIRST_VALUE = 'First'
    MIDDLE_VALUE = 'Middle'
    LAST_VALUE = 'Last'


def test_get_next_prev():
    '''Test enums.get_next and enums.get_prev'''
    assert DummyEnum.get_next(DummyEnum.FIRST_VALUE) \
        == DummyEnum.MIDDLE_VALUE
    assert DummyEnum.get_next(DummyEnum.MIDDLE_VALUE) \
        == DummyEnum.LAST_VALUE
    assert DummyEnum.get_next(DummyEnum.LAST_VALUE) \
        == DummyEnum.FIRST_VALUE
    assert DummyEnum.LAST_VALUE \
        == DummyEnum.get_prev(DummyEnum.FIRST_VALUE)


def test_from_int():
    '''Test enums can be converted from int.'''
    actual = None
    try:
        DummyEnum.from_int(12)
    except ValueError as ex:
        actual = ex
    assert actual


def test_repr():
    '''Test enums can be converted to text.'''
    assert f'{DummyEnum.FIRST_VALUE}' == 'First'



'''Module exporting a function to output a debug string.'''

from typing import Any

OUT_FUNC: Any = None


def out(*argv) -> None:
    '''Outputs the given arguments.'''
    if OUT_FUNC:
        OUT_FUNC(*argv)
